/**
 *
 * Copyright (c) 1996 Sun Microsystems, Inc.
 *
 * Use of this file and the system it is part of is constrained by the
 * file COPYRIGHT in the root directory of this system.
 *
 */

package COM.sun.labs.jjtree;

import java.util.Hashtable;

import COM.sun.labs.javacc.*;

public class JJTreeMain {

  static void help_message() {
    System.out.println("Usage is one of:");
    System.out.println("    jjtree option-settings - (to read from standard input)");
    System.out.println("OR");
    System.out.println("    jjtree option-settings inputfile (to read from a file)");
    System.out.println("");
    System.out.println("\"option-settings\" is a sequence of settings separated by spaces.");
    System.out.println("Each option setting must be of one of the following forms:");
    System.out.println("");
    System.out.println("    -optionname=value (e.g., -STATIC=false)");
    System.out.println("    -optionname:value (e.g., -STATIC:false)");
    System.out.println("    -optionname       (equivalent to -optionname=true.  e.g., -STATIC)");
    System.out.println("    -NOoptionname     (equivalent to -optionname=false. e.g., -NOSTATIC)");
    System.out.println("");
    System.out.println("Option settings are not case-sensitive, so one can say \"-nOsTaTiC\" instead");
    System.out.println("of \"-NOSTATIC\".  Option values must be appropriate for the corresponding");
    System.out.println("option, and must be either an integer or a boolean value.");
    System.out.println("");

    System.out.println("The integer valued options are:");
    System.out.println("");
    System.out.println("    NODE_STACK_SIZE  (default 500)");
    System.out.println("");
    System.out.println("The boolean valued options are:");
    System.out.println("");
    System.out.println("    STATIC                 (default true)");
    System.out.println("    MULTI                  (default false)");
    System.out.println("    NODE_DEFAULT_VOID      (default false)");
    System.out.println("    CHECK_DEFINITE_NODE    (default true)");
    System.out.println("");
    System.out.println("The string valued options are:");
    System.out.println("");
    System.out.println("    NODE_PREFIX            (default \"AST\")");
    System.out.println("    NODE_PACKAGE           (default \"\")");
    System.out.println("    OUTPUT_FILE            (default remove input file suffix, add .jj)");
    System.out.println("");
    System.out.println("JJTree also accepts JavaCC options, which it inserts into the generated file.");
    System.out.println("");

    System.out.println("EXAMPLES:");
    System.out.println("    jjtree -STATIC=false mygrammar.jjt");
    System.out.println("    jjtree - < mygrammar.jjt");
    System.out.println("");
    System.out.println("ABOUT JJTree:");
    System.out.println("    JJTree is a preprocessor for JavaCC built by Rob Duncan");
    System.out.println("    (rob.duncan@eng.sun.com) of Sun Microsystems Laboratories.");
    System.out.println("    To learn more about JavaCC, please visit the JavaCC web site at:");
    System.out.println("                http://www.suntest.com/JavaCC/");
    System.out.println("    If you wish to contact us, please send email to:");
    System.out.println("                javacc-help@asap.eng.sun.com");
    System.out.println("    If you wish to be added to the mailing list, please send email to:");
    System.out.println("                javacc-request@asap.eng.sun.com");
    System.out.println("    To post to the mailing list, please send email to:");
    System.out.println("                javacc-interest@asap.eng.sun.com");
  }

  static boolean isOption(String opt) {
    return opt.length() > 1 && opt.charAt(0) == '-';
  }

  /**
   * A main program that exercises the parser.
   */
  public static void main(String args[]) {

    JavaCCGlobals.bannerLine("Tree Builder", "0.2.5");

    JJTreeInit();
    JJTreeParser parser = null;
    if (args.length == 0) {
      System.out.println("");
      help_message();
      System.exit(1);
    } else {
      System.out.println("(type \"jjtree\" with no arguments for help)");
    }

    if (isOption(args[args.length-1])) {
      System.out.println("Last argument \"" + args[args.length-1] + "\" is not a filename or \"-\".  ");
      System.exit(1);
    }
    for (int arg = 0; arg < args.length-1; arg++) {
      if (!isOption(args[arg])) {
        System.out.println("Argument \"" + args[arg] + "\" must be an option setting.  ");
        System.exit(1);
      }
      Options.setCmdLineOption(args[arg]);
    }

    if (args[args.length-1].equals("-")) {
      System.out.println("Reading from standard input . . .");
      parser = new JJTreeParser(new java.io.DataInputStream(System.in));
    } else {
      System.out.println("Reading from file " + args[args.length-1] + " . . .");
      try {
	String arg = args[args.length-1];
        java.io.File fp = new java.io.File(arg);
        if (!fp.exists()) {
           System.out.println("File " + arg + " not found.");
           System.exit(1);
        }
        if (fp.isDirectory()) {
	  System.out.println(arg + " is a directory. Please use a valid file name.");
	  System.exit(1);
        }
	if (JavaCCGlobals.isGeneratedBy("JJTree", arg)) {
	  System.out.println(arg + " was generated by jjtree.  Cannot run jjtree again.");
	  System.exit(1);
	}
	JJTreeGlobals.toolList = JavaCCGlobals.getToolNames(arg);
	JJTreeGlobals.input_file = fp.getName();
        parser = new JJTreeParser(new java.io.DataInputStream(new java.io.FileInputStream(args[args.length-1])));
      } catch (NullPointerException ne) { // Should never happen
      } catch (SecurityException se) {
        System.out.println("Security violation while trying to open " + args[args.length-1]);
        System.exit(1);
      } catch (java.io.FileNotFoundException e) {
        System.out.println("File " + args[args.length-1] + " not found.");
        System.exit(1);
      }
    }
    try {

      parser.javacc_input();
      JJTree.start();

      if ((JavaCCErrors.get_error_count() == 0) && (Options.B("BUILD_PARSER") || Options.B("BUILD_TOKEN_MANAGER"))) {
        if (JavaCCErrors.get_warning_count() == 0) {
          System.out.println("Annotated grammar generated successfully in " + JJTreeGlobals.output_file);
        } else {
          System.out.println("Annotated grammar generated with 0 errors and "
                             + JavaCCErrors.get_warning_count() + " warnings.");
        }
        System.exit(0);
      } else {
        System.out.println("Detected " + JavaCCErrors.get_error_count() + " errors and "
                           + JavaCCErrors.get_warning_count() + " warnings.");
        System.exit((JavaCCErrors.get_error_count()==0)?0:1);
      }
    } catch (COM.sun.labs.javacc.MetaParseError e) {
      System.out.println("Detected " + JavaCCErrors.get_error_count() + " errors and "
                         + JavaCCErrors.get_warning_count() + " warnings.");
      System.exit(1);
    } catch (COM.sun.labs.javacc.ParseError e) {
      System.out.println("Detected " + (JavaCCErrors.get_error_count()+1) + " errors and "
                         + JavaCCErrors.get_warning_count() + " warnings.");
      System.exit(1);
    }
  }


  static Hashtable jjtreeOptions = new Hashtable();

  /**
   * Initialize for JJTree
   */
  public static void JJTreeInit() {

    Options.JavaCCInit();

    Options.optionValues.put("MULTI", Boolean.FALSE);
    Options.optionValues.put("NODE_PREFIX", "AST");
    Options.optionValues.put("NODE_PACKAGE", "");
    Options.optionValues.put("NODE_STACK_SIZE", new Integer(500));
    Options.optionValues.put("NODE_DEFAULT_VOID", Boolean.FALSE);
    Options.optionValues.put("OUTPUT_FILE", "");
    Options.optionValues.put("CHECK_DEFINITE_NODE", Boolean.TRUE);

    JJTreeGlobals.jjtreeOptions.put("MULTI", Boolean.FALSE);
    JJTreeGlobals.jjtreeOptions.put("NODE_PREFIX", "AST");
    JJTreeGlobals.jjtreeOptions.put("NODE_PACKAGE", "");
    JJTreeGlobals.jjtreeOptions.put("NODE_STACK_SIZE", new Integer(500));
    JJTreeGlobals.jjtreeOptions.put("NODE_DEFAULT_VOID", Boolean.FALSE);
    JJTreeGlobals.jjtreeOptions.put("OUTPUT_FILE", "");
    JJTreeGlobals.jjtreeOptions.put("CHECK_DEFINITE_NODE", Boolean.TRUE);
  }

}

