

/**
 * Static class with methods that return a string value casted
 * from another type. Cast to and from the same type acts as
 * a JRelix valid range checker.
 *
 * @since jRelCaNu 2002-07
 */


import java.util.*;


public class CastToStrn implements Constants, TypeConstants {  
    static private final String RESULT_TYPE_DC = STRING_DC;
    static private final String RESULT_TYPE_DK = STRING_DK;
    public static String assertValidValue(String castFromStrn) 
        throws InterpretError {
        return value(castFromStrn);
    };

    /**
     * Checks to see if one JRelix type is castable to another.
     * We must provide a value() method for each compatible type.
     *
     * @return  a boolean true if the RHS type can be cast to LHS type.
     * @since jRelCaNu
     */
    public static boolean domCompatible(int typeRHS) {
        switch (typeRHS) {
        case RELIX_VOID:
            return false;

        case BOOLEAN:
        case SHORT:
        case INTEGER:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case NUMERIC:
        case STRING:
        case TEXT:
        case DC:	//added by Yi Zheng Sept 28, 2002
        case DK:	//added by Yi Zheng Sept 28, 2002
        case ATTRIBUTE:  // moved here by Fan Guo on July 7, 2004	
        case UNIVERSAL:  // moved here by Fan Guo September 29, 2004
            return true;

        case STMT:
        case EXPR:
        case IDLIST:
        case COMP:
//        case DC:
//        case DK:
        case RELATION:
        case VIEW:
        case COMPUTATION:
      //  case ATTRIBUTE:	// moved up, return true by Fan
     //   case UNIVERSAL:   // moved up, return true by Fan
        default:
            return false;
        }
    }

    public static String value(String castFromStrn) {
        return castFromStrn;
    }

    public static String value(int castFromInt) 
        throws InterpretError {
        if (castFromInt == INT_DK)
            return STRING_DK;
        else if (castFromInt == INT_DC)
            return STRING_DC;
        else 
            return (new Integer(castFromInt).toString());
    }

    public static String value(short castFromShort) 
        throws InterpretError {
        if (castFromShort == SHORT_DK)
            return STRING_DK;
        else if (castFromShort == SHORT_DC)
            return STRING_DC;
        else 
            return (new Short(castFromShort).toString());
    }

    public static String value(long castFromLong) 
        throws InterpretError {
        if (castFromLong == LONG_DK)
            return STRING_DK;
        else if (castFromLong == LONG_DC)
            return STRING_DC;
        else 
            return (new Long(castFromLong).toString());
    }

    public static String value(byte castFromBool) 
        throws InterpretError {
        if (castFromBool == (byte) BOOL_DK)
            return STRING_DK;
        else if (castFromBool == (byte) BOOL_DC)
            return STRING_DC;
        else if (castFromBool == (byte) BOOL_FALSE) 
            return Boolean.FALSE.toString();
        else if (castFromBool == (byte) BOOL_TRUE) 
            return Boolean.TRUE.toString();
        else 
            throw new InterpretError(
                    "CastToStrn.value: error: Can't cast from unknown BOOL.");
    }

    public static String value(double castFromDouble) 
        throws InterpretError {
        try {
            if (castFromDouble == DOUBLE_DC)
                return STRING_DC;
            else if (castFromDouble == DOUBLE_DK)
                return STRING_DK;
            else if (Double.isInfinite(castFromDouble))
                throw new InterpretError(
                        "CastToStrn.value: error: Can't cast +/- Infinity.");
            else if (Double.isNaN(castFromDouble))
                throw new InterpretError(
                        "CastToStrn.value: error: Can't cast NaN.");
            else {
                return (new Double(castFromDouble).toString());
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToStrn.value(): error: Can't cast from castFromDouble:" + castFromDouble + " with JAVA cast with error: " + e);
        }
    }

    public static String value(float castFromFloat) 
        throws InterpretError {
        try {
            if (castFromFloat == FLOAT_DC)
                return STRING_DC;
            else if (castFromFloat == FLOAT_DK)
                return STRING_DK;
            else if (Float.isInfinite(castFromFloat))
                throw new InterpretError(
                        "CastToStrn.value: error: Can't cast +/- Infinity.");
            else if (Float.isNaN(castFromFloat))
                throw new InterpretError(
                        "CastToStrn.value: error: Can't cast NaN.");
            else {
                return (new Float(castFromFloat).toString());
            }
        } catch (Exception e) {
            throw new InterpretError(
                    "CastToStrn.value(): error: Can't cast from castFromFloat:" + castFromFloat + " with JAVA cast with error: " + e);
        }
    }

    public static String value(number castFromNumber) {
        String theNumber;
        String theExp;
        String theFraction;

        if (castFromNumber.iszero)
            return "+0e+000";

        byte[] fb = new byte[ castFromNumber.fractionLength ];

        for (int i = 0; i < castFromNumber.fractionLength; i++) 
            fb[i] = (byte) (castFromNumber.fraction[i] + 48);
        theFraction = new String(fb);
        if (castFromNumber.sign == '-')
            theFraction = "-." + theFraction;
        else
            theFraction = "+." + theFraction;
 
        int expVal = castFromNumber.expo > 0 ? castFromNumber.expo : -castFromNumber.expo;

        theExp = (new Integer(expVal)).toString();
        while (theExp.length() < 3)
            theExp = "0" + theExp;
        if (castFromNumber.expo < 0)
            theExp = "-" + theExp;
        else
            theExp = "+" + theExp;

        theNumber = theFraction + 'e' + theExp;
    
        return theNumber;

    }
}
