 /**
 * Class Computation.
 *
 * The constructor handles a computation declaration.
 * Method applyIjoin() handles the application of the computation 
 *    with a relation.
 * Method applySelect() handles the application of the computation
 *    with a select statement.
 *   -- applySelect(Relation rel,int currow ...) is overloaded for 
 *      implementing accessor methods of ADT
 * Method applyInOut() handles a stand-alone computation call
 *   -- applyInOut(Relation rel, ...) is overloaded for implementing
 *      modifier methods of ADT
 * Method duplicate() returns a duplicate of this Computation
 * Method print() is overloaded for the pr command on computation
 * Method sc() is for the sc command
 *
 * Notes on the use of Environments.
 * We use the discipline of 'static binding' (see Thesis).
 * 1) 'declEnv' stores a pointer to the environment in which the
 *    computation is declared (code is in constructor).
 * 2) 'declEnv' is used as the parent of the environment ('env') in which 
 *    the computation is applied (code is in method initializeEnvironment()
 *    which is called by apply* methods).
 *
 * @author Patrick Baker
 * @author Yi Zheng
 */
import java.util.*;
import java.io.*;

class Computation extends Relation 
implements Constants,TypeConstants
{  
   /***************************************************************/
   /*                           FIELDS                            */
   /***************************************************************/
	private boolean DEBUG=false;

	//added by Yi Zheng April 18, 2002 
	//stateEnv stores all state vars in a user session for this computation
	public Environment stateEnv;
	
	//added by Yi Zheng April 23, 2002 to distinguish between Ijoin and Select invocation
	public boolean isIjoin;
	
	//added by Yi Zheng May 14, 2002, to record the environment in which this comp is defined
	public String parentName = new String("");
	
	//added by Yi Zheng May 31, 2002, to distinguish between top level invocation by in/out syntax
	public boolean TOP_LEVEL_INOUT=false;
	
	//added by Yi Zheng June 3, 2002, to distinguish between calls made from within UPDATE 
	public boolean UPDATE = false;


   public CompBlock[] blocks;    // The "alt" blocks of the computation
   private Environment declEnv;  // Pointer to parent environment
   
   //Private Fields used during the application of this computation
   //REVISIT: pass these as parameters
   private Domain[] joinAttrs;  //The input attributes
   private Domain[] extraAttrs; //attrs that aren't part of calculation
   
   //This next field is set in the constructor of the class Interpreter
   //Actually it is set in JRelix.java --Yi Zheng

   public static Interpreter interpreter;

   /**
    * The constructor handles the declaration of new computations
    * @author Patrick Baker
    * modified by Weizhong Sun for event handler
    * modified by Yi Zheng for ADT
    */
   public Computation(SimpleNode node, Environment declEnv)
   throws InterpretError
   {  this.declEnv = declEnv; //store a pointer to the environment in which
						  //the declaration is made. This is used to 
						  //implement the discipline of 'static binding'.

		//REVISIT: Fix myEnv in class Relation.
		myEnv = declEnv; 
		
		//Yi Zheng April 18, 2002 to fix STATE bug
		stateEnv = new Environment(null); //no parent, don't load data
		//maybe we can load persistent state data here if necessary
		isIjoin = false; //April 23, 2002


	  //Dissect the parts of the computation from the syntax tree
	  SimpleNode compname; //The name of the computation
	  SimpleNode identifier; //The name of the computation is identifier which is the child of compname
	  SimpleNode parameters; //note: The parameter list could be empty
	  SimpleNode body;       //The declaration section + the 'alt' blocks
	  SimpleNode eventnode;	// the name for event
	  SimpleNode attrlist;	// the attribute list for event
          int active;       // the status of event handler

	  if (node.jjtGetNumChildren() == 3)
	  {  body       = (SimpleNode) node.jjtGetChild(2);
		 parameters = (SimpleNode) node.jjtGetChild(1);
		 compname = (SimpleNode) node.jjtGetChild(0);
	  }
	  else if (node.jjtGetNumChildren() == 2)
	  {  //Empty parameter list in this case
		 body       = (SimpleNode) node.jjtGetChild(1);
		 parameters = null;
		 compname = (SimpleNode) node.jjtGetChild(0);
	  }
	  else
	  {  //The Parser should really prevent this from ever happening
		 throw new 
		 InterpretError("Error: Incorrect computation definition");
	  }
	  

	  //Make sure all the parameters of the computation exist as domains
	  if (parameters != null)
	  {  int numChildren = parameters.jjtGetNumChildren();
		 domains = new Domain[numChildren]; 
		 for(int i=0; i<numChildren; i++)
		 {  String paramName = ((SimpleNode) parameters.jjtGetChild(i)).name;
			Domain d = declEnv.lookupDom(paramName, true);
			if( d == null)
			{  throw new 
			   InterpretError("Domain "+paramName+" does not exist");
			}
			domains[i] = d;
		 }
	  }
	  else
	  {  domains = new Domain[0];
	  }
	  
	  
	  //Initialize entries for .rel system table
	if (compname.opcode == OP_COMPIDENTIFIER) {
	  identifier = (SimpleNode) compname.jjtGetChild(0);
	  name = identifier.name;
	} else {
	// event
	  if (compname.jjtGetNumChildren() == 1) {
		 eventnode = (SimpleNode) compname.jjtGetChild(0);
		 attrlist = null;
	  } else {
		// having attribute list node
		 eventnode = (SimpleNode) compname.jjtGetChild(0);
		 attrlist = (SimpleNode) compname.jjtGetChild(1);
	  } 

	  name = declEnv.getEventName(eventnode);
	  
	  // add attrlist to name
	  if (attrlist != null) {
		int childnum = attrlist.jjtGetNumChildren();
		SimpleNode child;
		if (childnum != 0) {
			child = (SimpleNode)attrlist.jjtGetChild(0);
			name = name + "[" + child.name;
			for (int i = 1; i < childnum; i++) {
				child = (SimpleNode)attrlist.jjtGetChild(i);
				name = name + "," +child.name;
			}
			name = name + "]";
		}
	  }
          this.active = node.active;
	}

	  rvc = COMPUTATION;
	  numtuples = 0;
	  numattrs = (parameters == null) ? 0
									  : parameters.jjtGetNumChildren();
	  numsortattrs = 0;
	  tree = node;
	  
	  //Process the code sections now (i.e. "alt" blocks)
	  //Store computation code in array blocks[]
	  int numChildren = body.jjtGetNumChildren();
	  if (numChildren <= 0)
	  {  //The Parser should really prevent this from ever happening
		 throw new 
		 InterpretError("Error: Incorrect computation definition\n"
					  + "Computation has no body");
	  }
	  //use an offset of 1 if the first child of computation body is
	  //an optional declaration section and not an "alt" block
	  SimpleNode firstChild = (SimpleNode) body.jjtGetChild(0);
	  int offset = (firstChild.type == OP_COMPVARDECL) ? 1 : 0;
	  blocks = new CompBlock[numChildren-offset];
	  

	  for(int i=offset; i<numChildren; i++)
	  { Environment env = new Environment(declEnv);//each 'alt' block gets its own environment Yi Zheng
		initializeEnvironment(tree, env, null);//process parameters and local/state var decls
		try{
		 	blocks[i-offset] = 
		 		new CompBlock((SimpleNode) body.jjtGetChild(i), env, tree);
			} catch (SemanticCheckError sce)
			{ throw new InterpretError("semantic error");
			}
	  }
   } /* Constructor */   
 
   /***************************************************************/
   /*                          COMMANDS                           */
   /***************************************************************/
   /**
	* This method overides the print method of the Relation class.
	* @author Patrick Baker
	*/
   public void print(int[] lens, boolean flag)
   {  FileInputStream fin = null;
	  DataInputStream din = null;
	  String line;
	  try 
	  {  fin = new FileInputStream(name + ".comp");
		 din = new DataInputStream(fin);
		 while (true) 
		 {  line = din.readLine();
			if (line != null) 
			{  Global.println(line);
			}
			else 
			{  try 
			   {  if (fin != null) fin.close();
				  if (din != null) din.close();
			   } 
			   catch (IOException e) 
			   {}
			 //  return; modified by Yi Zheng March 19, 2002
			 break;
			}
		 }       
	  } 
	  catch (FileNotFoundException e) 
	  {  Global.println("File for computation " + name + " not found");
	  } 
	  catch (IOException e) 
	  {  Global.println("IOException in reading the file for computation " + name);
	  } 
	  finally 
	  {  try 
		 {  if (fin != null) fin.close();
			if (din != null) din.close();
		 } 
		 catch (IOException e) 
		 {}
	  }
   }   /* print() */ 

 	/**
 	 * handles the sc system command
 	 * @author Patrick Baker
 	 */
   public void sc()
   {  Vector inputs = new Vector();  //will hold the input parameters
	  Vector outputs = new Vector(); //will hold the output parameters

	  //Print out the computation name and the list of parameters
	  Global.print(name);
	  Global.print(" (");
	  for(int i=0; i<domains.length-1; i++)
	  {  Global.print(domains[i].name+", ");
	  }
	  if(domains.length > 0)
	  {  Global.print(domains[domains.length-1].name);
	  }
	  Global.println(")");

	  //Analyze each 'alt' block of the computation
	  for(int i=0; i<blocks.length; i++)
	  {  inputs.removeAllElements();  //reset at each block
		 outputs.removeAllElements();

		 //Deduce the input and output parameters for  block 'i'
		 int type = blocks[i].blockType;
		 for(int j=0; j<domains.length; j++)
		 {  //If the bit j of 'type' is set then we have an input parameter
			if ((type & (1 << j)) > 0)
			{  inputs.addElement(domains[j].name);
			}
			else
			{  outputs.addElement(domains[j].name);
			}
		 }

		 //Now display the block type
		 Global.print("  [ ");
		 for(int j=0; j<inputs.size(); j++)
		 {  Global.print(inputs.elementAt(j)+" ");
		 }
		 Global.print("] ");
		 Global.print("-> ");
		 Global.print("[ ");
		 for(int j=0; j<outputs.size(); j++)
		 {  Global.print(outputs.elementAt(j)+" ");
		 }
		 Global.println("] ");
	  }
   }   /* sc() */
   
   /***************************************************************/
   /*             PUBLIC INVOCATION METHODS  (top level)          */
   /***************************************************************/

   /**
	* Apply this computation to a relation by means of an ijoin call
	* Handles named attribute join
	* @author Patrick Baker
	* modified by Yi Zheng for bug fixes
	*/
   public Relation applyIjoin(Relation rel, Domain[] domsc, Domain[] domsr,
						 String filename, Environment callEnv)
   throws InterpretError
   {  
     //added by Yi Zheng June 20, 2002, to handle mappings between domsc and domsr
     Hashtable nameMapping = new Hashtable(); //domsc.name -> domsr
     Hashtable revNameMapping = new Hashtable(); //domsr.name -> domsc
     //end of addition June 20, 2002
   	 isIjoin = true;	//Yi Zheng April 23, 2002
   	 
      //Create new environment structure to keep info about params & vars..
	  //This implements the discipline of static binding: The environment
	  //that the computation runs in ('env') has as its parent the 
	  //environment in which the computation was declared in ('declEnv').
	  Environment env = new Environment(declEnv);

	  //keep a backup copy of domains since this will be modified when
	  //the computation is applied, and will then have to be restored.
	  Domain[] originalDomains = domains; 

	  if (domsc==null  &&  domsr==null)
	  {  //Find the join attributes
		 joinAttrs = commonDomains(domains, rel.domains);

		 //extra attributes are those from the relation that do not 
		 //participate in the computation, because they are not 
		 //parameters of the computation
		 extraAttrs = extraDomains(domains, rel.domains);
	  }
	  else if(domsc!=null  &&  domsr!=null  &&  domsc.length==domsr.length)
	  { 
		//added by Yi Zheng June 20, 2002
		
			//check if relations contain valid join attributes
			 if (!isDomainsIn(domsc, domains)){
			   throw new InterpretError("invalid domains in ijoin with " + name);
			 }
		
			 if (!isDomainsIn(domsr, rel.domains))
			   throw new InterpretError("invalid domains in ijoin with " + rel.name);
		
			 if (existDuplicateDomain(domsc) || existDuplicateDomain(domsr))
			   throw new InterpretError("duplicate attributes in ijoin");
			
			
			//check for type mismatch
			for(int i = 0; i < domsc.length; i++)
			{
				if(domsc[i].type != domsr[i].type)
					throw new InterpretError("type mismatch between join attributes: " +
					domsc[i].name + " and " + domsr[i].name);
					
			  // if the joined on nested domains, the nested domains must
			  // defined on same set of attributes
		
				if (domsc[i].type == IDLIST && domsr[i].type == IDLIST
					&& (domsc[i].name.compareTo(".id") != 0) && (domsr[i].name.compareTo(".id") != 0)) 
				 {
					Relation l = env.lookupRel("." + domsc[i].name, true);
					Relation r = env.lookupRel("." + domsr[i].name, true);
					if (!l.isDomainsIn(r.domains) || !r.isDomainsIn(l.domains))
					  throw new InterpretError(domsc[i].name + " and " 
						 + domsr[i].name + " not defined on same set of attributes");
			
					// A further and subtle restriction is that the order of the
					// attributes on which the two nested join attributes are
					// defined must be the same. This restriction is due the
					// sort order implicitly on the two nested join attributes
					// produced by the heapSort before the join process. It
					// is quite difficult to release this restriction if not
					// impossible.
					for (int j = 0; j < l.domains.length; j++) {
					  if (l.domains[j] != r.domains[j])
						 throw new InterpretError(domsc[i].name + " and "
					   + domsr[i].name + " not have same order of attributes");
					}
				 }
			}
			
			Domain[] commonDom = commonDomains(domains, rel.domains);
			// check if relations are joined on all common attributes 
			if (!isDomainsIn(commonDom, domsc) || !isDomainsIn(commonDom, domsr))
			 throw new 
			 InterpretError("relations not ijoined on all common attributes");
			for (int i = 0; i < domsc.length; i++) {
			  if (isDomainIn(domsc[i], commonDom) && domsc[i] != domsr[i])
				throw new 
				 InterpretError("join attribute mismatch between " + domsc[i].name +
				  " and " + domsr[i].name);
			}
	
			//joinAttrs and extraAttrs
			joinAttrs = domsc;
			for(int k = 0; k < domsc.length; k++)
			{
				nameMapping.put(domsc[k].name, domsr[k]);   
				revNameMapping.put(domsr[k].name, domsc[k]);
			}
			extraAttrs = extraDomains2(domains, rel.domains, revNameMapping);   
			
		//end of Yi's addition June 20, 2002, for handling join with named attribute
	  } 
	  else
	  {  throw new 
		 InterpretError("Number of join attributes specified do not match");
	  }
	  
	  //use a duplicate of the tree, June 21, 2002
	  SimpleNode tempTree = (SimpleNode)tree.jjtDeepDuplicate();
	  
	  //added by Yi Zheng June 21, 2002
	  Domain[] tempJoinAttrs = joinAttrs;
	  if(domsc != null && domsr != null)//to enforce pass by name
	  {
	  	renameParamsInTree(tempTree, nameMapping); //replace input parameter name with argument name
	  //change tempJoinAttrs to be used by findBlock
	  	tempJoinAttrs = replaceJoinAttrs(joinAttrs, nameMapping);
	   
	    reGenerateBlocks(tempTree); 	//domains and blocks are regenerated						
	  }
	  //end of addition June 21, 2002
	  
	  initializeEnvironment(tempTree, env, callEnv); //moved here June 20, 2002

	  //Find the right "alt"block and run the computation block
	  CompBlock block = findBlock(tempJoinAttrs, env); //June 21, 2002
  
      //Load the data for rel, including its dot relations
      rel.loadAll(rel);
      initializeData(rel, env, tempJoinAttrs); //Initialize the data array, 
      rel.unload();
      env.data = data;
      data = null;
      
      //added by Yi Zheng April 14, 2002 to fix recursive computation bug
      int oldNumTuples = numtuples;
      run(block, env);

      //added by Yi Zheng April 14, 2002 to fix recursive computation bug
	  numtuples = oldNumTuples;

	  //Finished running computation. return the result
	  Relation result = makeRelation(filename, env);
	  //set myEnv to callEnv in result Yi Zheng April 23, 2002
	  result.setEnv(callEnv);
	  
	  if(result.data == null)
		result.load(true);
		
	  //Check if we are running in constraint mode
	  if(constraintMode(tempJoinAttrs, env))
	  {  Global.println("Run in constraint mode");
		 Relation res = result.mujoin(OP_IJOIN, null, rel, null, filename);
		 domains = originalDomains;
	  	 numattrs = domains.length;	
	  	 rel.unloadAll(rel);
	  	 isIjoin = false;	//Yi Zheng April 23, 2002
	     if(domsc != null && domsr != null) //June 21, 2002
	     	reGenerateBlocks(tree); //this will restore domains and blocks
		 return res;
	  }
	  
	  //Restore the domains array and numattrs. These were altered
	  //while applying this computation and need to be restored.
	  domains = originalDomains;
	  numattrs = domains.length;
	  rel.unloadAll(rel);
	  
	  isIjoin = false; //Yi Zheng April 23, 2002
	  if(domsc != null && domsr != null) //June 21, 2002
	    reGenerateBlocks(tree); //this will restore domains and blocks

	  return result;
   } /* applyIjoin() */
   
   /**
	* Apply this computation using the in/out syntax.
	*
	* This is implemented by bringing in the input parameters from
	* the calling environment ('callEnv') and turning them into local
	* variables of the computations in the innermost frame ('env').
	* This job is accomplished by 'processCallParams()'. Since the
	* relational algebra routines use the file-system. (there is no 
	* in-RAM evaluator) the syntax tree of the block that will be
	* executed has its identifiers renamed. This is done by 'renameIDs()'.
	* The block is then executed. Finally, the output parameters,
	* which are all local variables, are copied out to the calling 
	* environment ('env'). This routine does not use parameters.
	* Inputs are made into locals, and outputs become locals on the fly.
	* @author Patrick Baker
	* modified by Yi Zheng for bug fixes
	*/
   //BUGS:
   // 1) Shouldn't have to write out all parameters, just outputs
   //    and input that are a part update or inc. assignment.
   //    Could mark all outputs in block algorithm.
   // 2) Biao reorders domains in join routine.
   // 3) Clean up files that are created by Biao's relational algebra code
   public void applyInOut(SimpleNode node, String filename,
							   Environment callEnv)
   throws InterpretError
   {
   	  isIjoin = false; //Yi Zheng April 23, 2002
   	  String prefix = "__#"+name;

	  //Create new environment structure to keep info about params & vars..
	  //This implements the discipline of static binding: The environment
	  //that the computation runs in ('env') has as its parent the 
	  //environment in which the computation was declared in ('declEnv').
	  Environment env = new Environment(declEnv);

	  //since the parameter list might be empty...
	  SimpleNode callParams = null; //holds the in/out parameters
	  if(node.jjtGetNumChildren() == 2) 
	  {  
	    callParams = (SimpleNode) node.jjtGetChild(1);
	  }
	  int blockType = processCallParams(callParams, prefix, env, callEnv);

	  //Find the right "alt" block and run the computation block
	  CompBlock block = null;
	  for(int i=0; i<blocks.length; i++)
	  { if (blocks[i].blockType == blockType)
		{  block = blocks[i];
		   break;
		}
	  }
	  if (block == null)
	  {  throw new InterpretError("No matching block type in computation");
	  }

	  
	  //Rename the identifiers in the tree for the block
	  //Use a duplicate of the tree.
	  SimpleNode newTree = block.blockTree.jjtDeepDuplicate(); 

//	  Environment tempEnv = new Environment(null);
// Yi Zheng May 15, 2002, changed to the following to support IDLIST typed state/local
	  Environment tempEnv = new Environment(env);

	  initializeEnvironment(tree, tempEnv, callEnv);//last argument changed from null, Yi Zheng, May 31, 2002
	  renameIDs(newTree, prefix, tempEnv);
	  
	  //Run the computation..
	  for(int stmtNo=0; stmtNo<newTree.jjtGetNumChildren(); stmtNo++)
	  {  SimpleNode stmt = (SimpleNode) newTree.jjtGetChild(stmtNo);

		//	 runSingleStmt(stmt, env);
		//modified by Yi Zheng May 17, 2002
		
		//addition by Yi Zheng May 31, 2002
		TOP_LEVEL_INOUT = true;
		//end of addition May 31, 2002
		
		runSingleStmt(stmt,tempEnv,callEnv); // last argument added by Zongyan Nov 25, 2004
			
		//addition by Yi Zheng May 31, 2002
		TOP_LEVEL_INOUT = false;
		//end of addition May 31, 2002
	
		 //REVISIT: delete any temp files...
	  }

	  //modified by Zongyan Nov 25, 2004
          //Tim Merrette: I'd say that if something (a relation) which is not a parameter
          //is changed in a computation, it should remain changed after the computation
          //exits.  

	  //now put all the parameters back into the call environment
	  for(int i=0; i<numattrs; i++)  // for each parameter
	  {  
		//REVISIT: should not need this if.
		//Problem is that biao reorders the domains in the join routine
		//so input relations can't be overwritten safely
		//REVISIT:only export relations in top-level env.
		if ( (blockType & (1 << i)) == 0 )  // only out parameters
		{
	   //	 IDInfo id = env.lookup(prefix+domains[i].name, false); modified by Yi Zheng May 17, 2002
		 IDInfo id = tempEnv.lookup(prefix+domains[i].name, false);
		 if (id == null)
		 {  //must be an output parameter that was never assigned to
			//try to catch this earlier as a semantic check:
			//All params are either in or out.
			//for now do nothing
		 }
		 else if(id.kind != REL) //in can be anything, but out can only be relations
		 {  //Only export relations.
			throw new InterpretError("Only relations can be exported; kind of "+prefix+domains[i].name+" is "+id.kind);
		 }
		 else
		 {  Relation r = ((RelInfo) id).rel;
			if (r == null)
			{  throw new InterpretError("Severe Error in applyInOut");
			}
	        // Bug fix by Yi Zheng March 25, 2002		String cpname = ((SimpleNode) callParams.jjtGetChild(i)).name;
			String cpname =((SimpleNode) ((SimpleNode) callParams.jjtGetChild(i)).jjtGetChild(0)).name;
			if (r.data == null)
			{  r.load();
			}
			r.name = cpname;
			
			//The order of the next two statements is important.
			//put() will call RelTable.addRel() which will remove the 
			//datafile on disk. Had we saved first, this would be lost.
		//	callEnv.put(r);
		//	r.save();
					// start Zongyan's addition Aug 18,2002
			SimpleNode cp = (SimpleNode) callParams.jjtGetChild(i);
			if (cp.url == null){
			    callEnv.put(r);
			    r.save();
			} else{
			    Relation rel = r;
			    rel.name = interpreter.nextTempName();
			    callEnv.put(rel);
			    rel.save();
		        interpreter.distributedStmt(rel.name,cpname,cp.url,OP_ASSIGN, callEnv);
		 	    callEnv.removeRel(rel.name, false);	
			}
			
			// end Zongyan's addition Aug 18.2002

		
		
		
		
		 }
		}
	  }
   } /* applyInOut() */   
   
    /**
	* Apply this computation to a relation using a selector
	* @author Patrick Baker
	* modified by Yi Zheng for bug fixes
	*/
   public Relation applySelect(SimpleNode select, String filename,
							   Environment callEnv)
   throws InterpretError
   { 
   	  isIjoin = false;	//Yi Zheng April 23, 2002
   	
      //Create new environment structure to keep info about params & vars..
	  //This implements the discipline of static binding: The environment
	  //that the computation runs in ('env') has as its parent the 
	  //environment in which the computation was declared in ('declEnv').
	  Environment env = new Environment(declEnv);
	  initializeEnvironment(tree, env, callEnv); //state dom shouldn't be exported here, Yi Zheng, May 30, 2002
	//as per conversation on May 30, 2002, state dom can still be exported
//		initializeEnvironment(tree, env, null);

	  createDataArray(1);
	  		
	  Vector inputDoms = new Vector(); //will hold the input domains
	  parseSelectConditional(select, inputDoms, env, callEnv); //CONTINUE MAY 30
	  numtuples = 1; //BUG: REVISIT: this could be wrong whenever there is ALSO clause
	  				// but unfortunately now ALSO is not correctly implented in the parser
	  				// Yi Zheng March 31, 2002
	  env.data = data;
	  data = null;
	  
	  //Copy the input domains from the vector 'inputdoms' to
	  //the array 'joinAttrs' for compatibility with the applyIjoin
	  //method. From then on, we reuse methods that applyIjoin 
	  //needed.
	  	
	  joinAttrs = new Domain[inputDoms.size()];
	  for(int i=0; i<inputDoms.size(); i++)
	  {  joinAttrs[i] = (Domain) inputDoms.elementAt(i);
	  }

	  load(this);
	  //Find the right "alt"block, determine if run in constraint mode, and run the computation block
	  CompBlock block = findBlock(joinAttrs, env);
	  
  	  //Check if we are running in constraint mode, Yi Zheng March 31, 2002
	  Vector failRows = new Vector(); //keeps track of which rows need to be kicked out
	  if (constraintMode(joinAttrs, env))
	  {  Global.println("Run in constraint mode");
	  	 runConstraint(block, env, failRows); //since env.data has data for all attributes, we just need to kick out those failing the constraint
		 if(failRows.size() == 0) failRows = null;
	  } 
	  else
	  { failRows = null; //for GC
	  	run(block,env);
	  }

	  
	  // run(block, env); Yi Zheng March 31, 2002
	  
	  //Finished running computation. return the result

	  //  Relation result = makeRelation(filename, env); Yi Zheng March 31, 2002
	  Relation result = null;
	  if(failRows == null)
		result = makeRelation(filename, env);
	  else result = makeRelationConstraint(filename, env, failRows);

	  //Yi Zheng April 23, 2002
	  result.setEnv(callEnv);	  
	  //Check if we are running in constraint mode
//	  if (constraintMode(joinAttrs, env))
//	  {  System.out.println("Run in constraint mode");
//	  } moved forward to after findBlock by Yi Zheng, March 31, 2002
	  unload(this);
	  return result;
   } /* applySelect */
 
   /***************************************************************/
   /*             PUBLIC INVOCATION METHODS  (nested level)       */
   /***************************************************************/
   /**
    * applyInOut, used for implementing modifier methods of ADT
    * @author Yi Zheng
    */
   //note that we shouldn't allow 'out' parameters in this nested case
   public Relation applyInOut(Relation rel, SimpleNode node, Environment callEnv) throws InterpretError
   {
   	UPDATE = true;
	isIjoin = false;	//Yi Zheng April 23, 2002
	String prefix = "__#"+name;
	Environment env = new Environment(declEnv);
	SimpleNode callParams = null;

	
	if (node.jjtGetNumChildren() == 2)
	{
		callParams = (SimpleNode) node.jjtGetChild(1);
	}
		
	int blockType = processCallParams(callParams, prefix, env, callEnv);

	int col;

	CompBlock block = null;
	for (int i=0; i<blocks.length; i++)
	{ 
	if (blocks[i].blockType == blockType)
		{ block = blocks[i];
		  break;
		}
	}
	if(block == null)
		throw new InterpretError("Block not found!");	
	SimpleNode newTree = block.blockTree.jjtDeepDuplicate();
	
//	Environment tempEnv = new Environment(null);
//  Yi Zheng May 15, 2002, changed to the following to support IDLIST state/local
	Environment tempEnv = new Environment(env);

	initializeEnvironment(tree, tempEnv, null); //REVISIT: check if the relVal in state/local is relevant

	renameIDs(newTree, prefix, tempEnv);
	
	int numStmts = newTree.jjtGetNumChildren();
	
	try{
		
		NREnvironment NRe = tempEnv.addNRFrame(rel);
		int nrows = rel.numtuples;
		for(NRe.row = 0; NRe.row < nrows; NRe.row++)
		{
			for (int j = 0; j < numStmts; j++)
			{ 
		  		SimpleNode stmt = ((SimpleNode) newTree.jjtGetChild(j)).jjtDeepDuplicate(); //stmt tree can be modified

		  		runSingleStmtInUpdate(rel, stmt, tempEnv, NRe);
		  	}
		}
		  
	}catch (InterpretError inte)
	{
	 UPDATE = false;
	 tempEnv.removeNRFrame();
	 throw inte;
	}
	
	UPDATE = false;
	tempEnv.removeNRFrame();
	rel.save(); //only when all stmts ran successfully will the changes to rel be final
    return rel; 
   } /* applyInOut(), nested */

   /**
	* Apply this computation to a relation using a selector
	*	where this computation is a nested attribute of rel
	* used to implement accessor method of ADT
	* @author Yi Zheng June 25, 2002
	*/
   public Relation applySelect(Relation rel, int currow, SimpleNode select, String filename,
							   Environment callEnv)
   throws InterpretError
   { 
   	  isIjoin = false;	//Yi Zheng April 23, 2002
   	
      //Create new environment structure to keep info about params & vars..
	  //This implements the discipline of static binding: The environment
	  //that the computation runs in ('env') has as its parent the 
	  //environment in which the computation was declared in ('declEnv').
	  Environment env = new Environment(declEnv);
	  NREnvironment compEnv = env.addNRFrame(rel);
	  compEnv.row = currow;
	  initializeEnvironment(tree, env, callEnv); //state dom shouldn't be exported here, Yi Zheng, May 30, 2002
	//as per conversation on May 30, 2002, state dom can still be exported
//		initializeEnvironment(tree, env, null);

	  createDataArray(1); //ALSO will crash this
	  		
	  Vector inputDoms = new Vector(); //will hold the input domains
	  
	  EvalExpr.WEIZHONG = true; //special treatment for NREnv
	  try{
	  	parseSelectConditional(select, inputDoms, env, callEnv); 
	  	EvalExpr.WEIZHONG = false;

	  }catch(InterpretError e)
	  { EvalExpr.WEIZHONG = false;
	    throw e;
	  }
	  numtuples = 1; //BUG: REVISIT: this could be wrong whenever there is ALSO clause
	  				// but unfortunately now ALSO is not correctly implented in the parser
	  				// Yi Zheng March 31, 2002
	  env.data = data;
	  data = null;
	  
	  //Copy the input domains from the vector 'inputdoms' to
	  //the array 'joinAttrs' for compatibility with the applyIjoin
	  //method. From then on, we reuse methods that applyIjoin 
	  //needed.
	  	
	  joinAttrs = new Domain[inputDoms.size()];
	  for(int i=0; i<inputDoms.size(); i++)
	  {  joinAttrs[i] = (Domain) inputDoms.elementAt(i);
	  }

	  load(this);
	  //Find the right "alt"block, determine if run in constraint mode, and run the computation block
	  CompBlock block = findBlock(joinAttrs, env);

	  //Check if we are running in constraint mode, Yi Zheng March 31, 2002
	  Vector failRows = new Vector(); //keeps track of which rows need to be kicked out
	  if (constraintMode(joinAttrs, env))
	  {  Global.println("Run in constraint mode");
	  	 EvalExpr.WEIZHONG = true;
	  	 try {
	  	 	runConstraint(block, env, failRows); //since env.data has data for all attributes, we just need to kick out those failing the constraint
			EvalExpr.WEIZHONG = false;
		 } catch (InterpretError e)
		 { EvalExpr.WEIZHONG = false;
		 	throw e;
		 }
		 if(failRows.size() == 0) failRows = null;
	  } 
	  else
	  { failRows = null; //for GC
	    EvalExpr.WEIZHONG = true;
	    try {
		  	run(block,env);
	  		EvalExpr.WEIZHONG = false;
	  	}catch (InterpretError e)
	  	{ EvalExpr.WEIZHONG = false;
	  	  throw e;
	  	}
	  }

	  
	  //Finished running computation. return the result
	  Relation result = null;
	  if(failRows == null)
		result = makeRelation(filename, env);
	  else result = makeRelationConstraint(filename, env, failRows);

		//Yi Zheng April 23, 2002
		result.setEnv(callEnv);	  
		
		unload(this);
		return result;
   } /* applySelect(), nested */
   
   /**
    * duplicate a computation
    * @author Yi Zheng
    */	
   public Computation duplicate()
   throws InterpretError
   {  return new Computation(((SimpleNode)tree).jjtDeepDuplicate(), declEnv);
   }   

   /*************************************************
    *   PRIVATE METHODS 							*
    *************************************************/
    
   /**
    * generate domains and blocks for this computation according to tempTree
    * @ author Yi Zheng June 21, 2002
    */
   private void reGenerateBlocks(SimpleNode tempTree)
   throws InterpretError
   {
	  SimpleNode identifier; //The name of the computation is identifier which is the child of compname
	  SimpleNode parameters; //note: The parameter list could be empty
	  SimpleNode body;       //The declaration section + the 'alt' blocks

	  if (tempTree.jjtGetNumChildren() == 3)
	  {  body       = (SimpleNode) tempTree.jjtGetChild(2);
		 parameters = (SimpleNode) tempTree.jjtGetChild(1);
	  }
	  else if (tempTree.jjtGetNumChildren() == 2)
	  {  //Empty parameter list in this case
		 body       = (SimpleNode) tempTree.jjtGetChild(1);
		 parameters = null;
	  }
	  else
	  {  //The Parser should really prevent this from ever happening
		 throw new 
		 InterpretError("Error: Incorrect computation definition");
	  }
	  

	  //Make sure all the parameters of the computation exist as domains

	  if (parameters != null)
	  {  int numChildren = parameters.jjtGetNumChildren();
		 domains = new Domain[numChildren]; 
		 for(int i=0; i<numChildren; i++)
		 {  String paramName = ((SimpleNode) parameters.jjtGetChild(i)).name;
			Domain d = declEnv.lookupDom(paramName, true);
			if( d == null)
			{  throw new 
			   InterpretError("Domain "+paramName+" does not exist");
			}
			domains[i] = d;
		 }
	  }
	  else
	  {  domains = new Domain[0];
	  }
	  
	  
	  numattrs = (parameters == null) ? 0
									  : parameters.jjtGetNumChildren();
	  
	  //Process the code sections now (i.e. "alt" blocks)
	  //Store computation code in array blocks[]
	  int numChildren = body.jjtGetNumChildren();
	  if (numChildren <= 0)
	  {  //The Parser should really prevent this from ever happening
		 throw new 
		 InterpretError("Error: Incorrect computation definition\n"
					  + "Computation has no body");
	  }
	  //use an offset of 1 if the first child of computation body is
	  //an optional declaration section and not an "alt" block
	  SimpleNode firstChild = (SimpleNode) body.jjtGetChild(0);
	  int offset = (firstChild.type == OP_COMPVARDECL) ? 1 : 0;
	  blocks = new CompBlock[numChildren-offset];
	  

	  for(int i=offset; i<numChildren; i++)
	  { Environment env = new Environment(declEnv);//each 'alt' block gets its own environment Yi Zheng
		initializeEnvironment(tempTree, env, null);//process parameters and local/state var decls
		try{
		 	blocks[i-offset] = 
		 		new CompBlock((SimpleNode) body.jjtGetChild(i), env, tempTree);
			} catch (SemanticCheckError sce)
			{ throw new InterpretError("semantic error");
			}
	  }
   } /* regenerateBlocks */
   
	/**
	 * invoked by transformExpr 
	 * return true if an IDInfo is of primitive type and instanceof
	 * StateInfo/LocalInfo
	 * @author Yi Zheng
	 */
	 private boolean isReplaceable(IDInfo id) throws InterpretError
	 {
	 	if (id == null) return false;
	 	switch(id.kind)
	 	{
	 	//	case PARAMETER: //since in applyInOut (top level), all param value is LocalInfo
	 		case STATE:
	 		case LOCAL:
	 			break;
	 		default:
	 			return false;
	 	}
	 	switch(id.type)
	 	{
	 		case BOOLEAN:
	 		case SHORT:
	 		case INTEGER:
	 		case LONG:
	 		case FLOAT:
	 		case DOUBLE:
	 //		case NUMERIC:
	 		case STRING:
	 			return true;
	 		case IDLIST:
	 		case COMP:
	 			return false;
	 		default:
	 			throw new InterpretError("Type not recognized");
	 	}
	 } /* isReplaceable */
	 
	/**
	 * recursively traverse the node and replace identifier 
	 * with a constant node if such identifier is 
	 * a primitive typed State/Local
	 * assumption: 1) no two local/state have the same name
	 *			   2) virtual domains have already been actualized
	 * @author Yi Zheng May 17, 2002
	 */
	 private void transformStmt(SimpleNode stmt, Environment env) throws InterpretError
	 {
	  	switch(stmt.type)
	 	{
	 		case OP_ASSIGNMENT:{
	 			int numChild = stmt.jjtGetNumChildren();
	 			SimpleNode rnode = null;
	 			if(numChild == 2)
	 			{
	 				rnode = (SimpleNode) stmt.jjtGetChild(1);
	 			}else if(numChild == 4)
	 			{
	 				rnode = (SimpleNode) stmt.jjtGetChild(3);
	 			}else throw new InterpretError("invalid assignment statement");
	 			if (rnode != null)
	 				transformExpr(rnode, env);
	 			break;
	 		}
	 		case OP_COMMAND:
	 			break;	//not likely used in update change stmtlist
	 		case OP_DECLARATION:
	 			break; //ditto
	 		case OP_STATEMENT:{
			   switch(stmt.opcode) 
			   {  case OP_CONDITIONAL :
					 SimpleNode ifNode, thenNode, elseNode = null;
					 ifNode = (SimpleNode) stmt.jjtGetChild(0);
					 thenNode = (SimpleNode) stmt.jjtGetChild(1);
			 		 
			 		 transformStmt(ifNode, env);
			 		 transformStmt(thenNode, env);
					 
					 if(stmt.jjtGetNumChildren() == 3)
					 {  elseNode = (SimpleNode) stmt.jjtGetChild(2);
						transformStmt(elseNode, env);
					 }
					 break;
				  case OP_STMTBLOCK :
					 for(int stmtNo=0; stmtNo<stmt.jjtGetNumChildren(); stmtNo++)
					 {  transformStmt((SimpleNode) stmt.jjtGetChild(stmtNo), env);
					 }
					 break;
				  case OP_COMPCALL: 
				  	 if(stmt.jjtGetNumChildren() > 1)
				  	 {
				  	 	SimpleNode paramList = (SimpleNode) stmt.jjtGetChild(1);
				  	 	for(int i = 0; i < paramList.jjtGetNumChildren(); i++)
				  	 	{
				  	 		SimpleNode param = (SimpleNode)paramList.jjtGetChild(i);
				  	 		if(!param.isBitSet(0)) //input param
				  	 			transformExpr((SimpleNode)param.jjtGetChild(0), env);
				  	 	}	
				  	 	
				  	 }
				  	break;
				  case OP_SEQUENCE :
				  case OP_PARALLEL :
				  case OP_CHOICE :
				  case OP_FORLOOP :
				  case OP_WHILELOOP :
				  case OP_EXIT :
				  case OP_DEADLOCK :
				  case OP_EXEC :
					 throw new 
					 InterpretError("Not implemented yet");
			   }
			   break;
	 			
	 		}
	 		case OP_UPDATE:{
				switch(stmt.opcode)
				{
					case OP_CHANGE:{
						SimpleNode stmtList = (stmt.isBitSet(0)?(SimpleNode)stmt.jjtGetChild(1):null);
						SimpleNode using = null;
						if(stmtList != null && stmt.isBitSet(1))
							using = (SimpleNode)stmt.jjtGetChild(2);
						else if(stmt.isBitSet(1))
							using = (SimpleNode)stmt.jjtGetChild(1); //impossible
						if(stmtList != null)
						{
							for(int i = 0; i < stmtList.jjtGetNumChildren(); i++)
								transformStmt((SimpleNode)stmtList.jjtGetChild(i), env);
						}
						if(using != null)
						{
							int numChild = using.jjtGetNumChildren();
							if(numChild == 1)
							{
								transformExpr((SimpleNode)using.jjtGetChild(0), env);
							}
							else if(numChild == 3)
							{
								transformExpr((SimpleNode)using.jjtGetChild(2), env);
							}
							else throw new InterpretError("invalid update change");
						}
						break;
					}
					case OP_ADD:
					case OP_DELETE:{
						int numChild = stmt.jjtGetNumChildren();
						if(numChild == 2)
						{
							transformExpr((SimpleNode)stmt.jjtGetChild(1), env);
						}
						else if(numChild == 4)
						{
							transformExpr((SimpleNode)stmt.jjtGetChild(3), env);
						}
						else throw new InterpretError("invalid update add/delete");
						break;
					}
					default:
						throw new InterpretError("invalid update type");
				}
	 			break;
	 		}
	 		default:
	 			throw new InterpretError("invalid statement type");
	 	}
	 
	 } /* transformStmt */
	 	
	
	 /**
	 * recursively traverse the node and replace identifier 
	 * with a constant node if such identifier is 
	 * a primitive typed State/Local
	 * assumption: 1) no two local/state have the same name
	 *			   2) virtual domains have already been actualized
	 * @author Yi Zheng May 17, 2002
	 */
	 private void transformExpr(SimpleNode node, Environment env) throws InterpretError
	 {
	 	if (node == null) return;
	 	
	   	int type = node.type;
	 	int opcode = node.opcode;
	  	String name = node.name;
	  	IDInfo id = null;
	  	Object obj = null;
	  	switch (type)
	  	{
	  		case OP_LITERAL:
	  			return;
	  	    case OP_IDENTIFIER :
	//	    case OP_EVENTNAME :
		    {
		      	id = env.lookup(name,true);
		      	if(isReplaceable(id))
		      	{
		      		 switch (id.type)
		      		 {
		      		 	case BOOLEAN:
		      		 		if(id.kind == LOCAL)
		      		 		{
		      		 			byte bt = (byte)(((LocalInfo)id).intVal);
		      		 			obj = new Byte(bt);
		      		 		}
		      		 		else //must be STATE
		      		 		{
		      		 			byte bt = (byte)(((StateInfo)id).intVal);
		      		 			obj = new Byte(bt);
		      		 		}
		      		 		break;
		      		 	case SHORT:
		      		 		if(id.kind == LOCAL)
		      		 		{
		      		 			short st = (short)(((LocalInfo)id).intVal);
		      		 			obj = new Short(st);
		      		 		}
		      		 		else //must be STATE
		      		 		{
		      		 			short st = (short)(((StateInfo)id).intVal);
		      		 			obj = new Short(st);
		      		 		}
		      		 		break;
		      		 	case INTEGER:
		      		 		if(id.kind == LOCAL)
		      		 		{
		      		 			int it = ((LocalInfo)id).intVal;
		      		 			obj = new Integer(it);
		      		 		}
		      		 		else //must be STATE
		      		 		{
		      		 			int it = ((StateInfo)id).intVal;
		      		 			obj = new Integer(it);
		      		 		}
		      		 		break;
		      		 	case LONG:
		      		 		if(id.kind == LOCAL)
		      		 		{
		      		 			long lt = ((LocalInfo)id).longVal;
		      		 			obj = new Long(lt);
		      		 		}
		      		 		else //must be STATE
		      		 		{
		      		 			long lt = ((StateInfo)id).longVal;
		      		 			obj = new Long(lt);
		      		 		}
		      		 		break;
		      		 	case FLOAT:
		      		 		if(id.kind == LOCAL)
		      		 		{
		      		 			float ft = ((LocalInfo)id).floatVal;
		      		 			obj = new Float(ft);
		      		 		}
		      		 		else //must be STATE
		      		 		{
		      		 			float ft = ((StateInfo)id).floatVal;
		      		 			obj = new Float(ft);
		      		 		}
		      		 		break;
		      		 	case DOUBLE:
		      		 		if(id.kind == LOCAL)
		      		 		{
		      		 			double dt = ((LocalInfo)id).doubleVal;
		      		 			obj = new Double(dt);
		      		 		}
		      		 		else //must be STATE
		      		 		{
		      		 			double dt = ((StateInfo)id).doubleVal;
		      		 			obj = new Double(dt);
		      		 		}
		      		 		break;
		      		 	case STRING:	
		      		 		if(id.kind == LOCAL)
		      		 		{
		      		 			String st = ((LocalInfo)id).stringVal;
		      		 			obj = new String(st);
		      		 		}
		      		 		else //must be STATE
		      		 		{
		      		 			String st = ((StateInfo)id).stringVal;
		      		 			obj = new String(st);
		      		 		}
		      		 		break;
		      		 	default:
		      		 		throw new InterpretError("Unrecognized type");

		      		 }
		      	
		      	
		      	}
		      	else return;
		      	
		      	//construct and replace node with newConstNode
		      	SimpleNode newConstNode = 
		      		SimpleNode.jjtCreateNode("Literal", OP_LITERAL, id.type + OP_BOOLEAN, name);
		      	newConstNode.setObject(obj);
		      	newConstNode.jjtSetParent(node.jjtGetParent());
		      	node.jjtReplace(newConstNode);
		      	return;
		    }
		    case OP_JOINOPERATOR : {
		      if(node.jjtGetNumChildren() == 2)
		      {
		      	transformExpr((SimpleNode)node.jjtGetChild(0),env);
		      	transformExpr((SimpleNode)node.jjtGetChild(1),env);
		      }
		      else //numchildren == 4
		      {
		      	transformExpr((SimpleNode)node.jjtGetChild(0),env);
		      	transformExpr((SimpleNode)node.jjtGetChild(1),env);
		      	transformExpr((SimpleNode)node.jjtGetChild(2),env);
		      	transformExpr((SimpleNode)node.jjtGetChild(3),env);
		      }
		      return;
		    }
		    case OP_EXPRLIST: {
		    	int numchildren = node.jjtGetNumChildren();
		    	for(int i = 0; i < numchildren; i++)
		    	{
		    		transformExpr((SimpleNode)node.jjtGetChild(i),env);
		    	}
		    	return;
		    }
		    case OP_PROJECT : {
			    // get the projector
			    SimpleNode nodel = (SimpleNode) node.jjtGetChild(0);
			    //get the selection
			    SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
				//REVISIT: do we only care about the selection?
				transformExpr(noder, env);
				return;
		    }
		    case OP_TSELECT : { 
		      // map this node to : projection + selection
		      node.type = OP_PROJECT;
		      node.opcode = OP_WHENIN;
		
		      transformExpr(node,env);
		      return;
		    }
		    case OP_SELECT : {
			    // get conditional expression for selection
			    SimpleNode nodel = (SimpleNode) node.jjtGetChild(0);
			    // get relational expression for selection
			    SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
				transformExpr(nodel, env);
				transformExpr(noder, env);
				return;
		    }
		    case OP_IFTHENELSE: {
		    	if (node.jjtGetNumChildren() != 3)
		    		throw new InterpretError("Wrong number of children of IF THEN ELSE");
		    	SimpleNode node_if = (SimpleNode) node.jjtGetChild(0);
		    	SimpleNode node_then = (SimpleNode) node.jjtGetChild(1);
		    	SimpleNode node_else = (SimpleNode) node.jjtGetChild(1);
		    	transformExpr(node_if, env);
		    	transformExpr(node_then, env);
		    	transformExpr(node_else, env);
		    	return;
		    }
		    case OP_QSELECT : {
				SimpleNode node_QTlist = (SimpleNode) node.jjtGetChild(0);
				SimpleNode node_cond = null;
				SimpleNode node_rel = null;
				if(node.jjtGetNumChildren() == 2)
				{
					node_rel = (SimpleNode) node.jjtGetChild(1);
				}
				else
				{
					node_cond = (SimpleNode) node.jjtGetChild(1);
					node_rel = (SimpleNode) node.jjtGetChild(2);
				}
				transformExpr(node_QTlist, env);
				transformExpr(node_rel, env);
				if (node_cond != null) transformExpr(node_cond, env);
				return;
		    }
		    case OP_QTLIST: {
		    	for (int i = 0; i < node.jjtGetNumChildren(); i++)
		    	{
		    		transformExpr((SimpleNode)node.jjtGetChild(i), env);
		    	}
		    	return;
		    }
		    case OP_QUANTIFIER: {
		    	transformExpr((SimpleNode)node.jjtGetChild(0), env);
		    	transformExpr((SimpleNode)node.jjtGetChild(1), env);
		    	return;
		    }
		    case OP_PICK : {
			    SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
				transformExpr(nd, env);
				return;
		    }
		    case OP_ARRAY :{  
				SimpleNode Identifier = 
				  (SimpleNode) node.jjtGetChild(0);
				SimpleNode arrayList = (SimpleNode) node.jjtGetChild(1);
			
				int ALindex=0; //index into the children of ArrayList
				for(int i=0; i<arrayList.jjtGetNumChildren(); i++)
				  {  if (arrayList.isBitSet(i))
				    {  
				      SimpleNode nd = (SimpleNode) arrayList.jjtGetChild(ALindex);
				      transformExpr(nd, env);
				      ALindex++;
				    }
				  }
				  return;
		      }
		    case OP_BIOPERATOR : {
			 	SimpleNode nodel = (SimpleNode) node.jjtGetChild(0);
			 	SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
			 	transformExpr(nodel, env);
			 	transformExpr(noder, env);
			 	
		      	return;
		    }
		    case OP_UOPERATOR: {
		    	transformExpr((SimpleNode) node.jjtGetChild(0), env);
		    	return;
		    }
		    case OP_FUNCTION: {
		    	transformExpr((SimpleNode) node.jjtGetChild(0), env);
		    	return;
		    }
		    case OP_RED: {
		    	if(node.opcode != OP_REDFUNCALL)
		    		transformExpr((SimpleNode) node.jjtGetChild(0), env);
		    	else
		    		transformExpr((SimpleNode) node.jjtGetChild(1), env);
		    		
		    	return;
		    }
		    case OP_FUN:
		    case OP_EQUIV: {
		    	if(node.opcode != OP_REDFUNCALL)
		    		transformExpr((SimpleNode) node.jjtGetChild(0), env);
		    	else
		    		transformExpr((SimpleNode) node.jjtGetChild(1), env);
		    		
		    	SimpleNode byList = null;
		    	if(node.opcode != OP_REDFUNCALL)
		    		byList = (SimpleNode) node.jjtGetChild(1);
		    	else
		    		byList = (SimpleNode) node.jjtGetChild(2);
		    	
		    	for(int i = 0; i < byList.jjtGetNumChildren(); i++)
		    		transformExpr((SimpleNode) byList.jjtGetChild(i), env);
		    	return;
		    }
		    case OP_PAR: {
		    	if(node.opcode != OP_REDFUNCALL)
		    		transformExpr((SimpleNode) node.jjtGetChild(0), env);
		    	else
		    		transformExpr((SimpleNode) node.jjtGetChild(1), env);
		    	SimpleNode byList = null;
		    	if(node.opcode != OP_REDFUNCALL)
		    		byList = (SimpleNode) node.jjtGetChild(1);
		    	else
		    		byList = (SimpleNode) node.jjtGetChild(2);
		    	for(int i = 0; i < byList.jjtGetNumChildren(); i++)
		    		transformExpr((SimpleNode) byList.jjtGetChild(i), env);
		    	SimpleNode orderList = null;
		    	if(node.opcode != OP_REDFUNCALL)
		    		orderList = (SimpleNode) node.jjtGetChild(2);
		    	else
		    		orderList = (SimpleNode) node.jjtGetChild(3);
		    		
		    	for(int i = 0; i < orderList.jjtGetNumChildren(); i++)
		    		transformExpr((SimpleNode) orderList.jjtGetChild(i), env);
		    	return;
		    }
		    default : {
		      throw new 
			InterpretError("invalid/not implemented relational expression");
		    }
	  	}
	 } /* transformExpr */
	 
/* original code from Weizhong
   // weizhong: override the applyInOut for update..comp...
   public Relation applyInOut(Relation rel, SimpleNode node, Environment callEnv) throws InterpretError
   {
	isIjoin = false;	//Yi Zheng April 23, 2002
	String prefix = "__#"+name;
	Environment env = new Environment(declEnv);
	SimpleNode callParams = null;

	if (node.jjtGetNumChildren() == 2)
	{
		callParams = (SimpleNode) node.jjtGetChild(1);
	}
		
	int blockType = processCallParams(callParams, prefix, env, callEnv);

	int col;

	CompBlock block = null;
	for (int i=0; i<blocks.length; i++)
	{ if (blocks[i].blockType == blockType)
		{ block = blocks[i];
		  break;
		}
	}
	if(block == null)
		throw new InterpretError("Block not found!");	
	SimpleNode newTree = block.blockTree.jjtDeepDuplicate();
//	Environment tempEnv = new Environment(null);
//  Yi Zheng May 15, 2002, changed to the following to support IDLIST state/local
	Environment tempEnv = new Environment(env);
	
	initializeEnvironment(tree, tempEnv, null); //REVISIT: check if the relVal in state/local is relevant
	renameIDs(newTree, prefix, tempEnv);
	
	//Yi Zheng May 14, 2002
        //tempEnv.setParent(env);

	int numStmts = newTree.jjtGetNumChildren();
	for (int j = 0; j < numStmts; j++)
	{ 
	  SimpleNode stmt = (SimpleNode) newTree.jjtGetChild(j);
	  
	  switch(stmt.type)
	  { case OP_ASSIGNMENT :
	      if (stmt.jjtGetNumChildren() == 2) 
	      { 
	      	SimpleNode left = (SimpleNode) stmt.jjtGetChild(0);
		SimpleNode expr = (SimpleNode) stmt.jjtGetChild(1);

		IDInfo idleft = env.lookup(left.name, true); 
 //		IDInfo idleft = tempEnv.lookup(left.name, true);
		
		//IDInfo idexpr = env.lookup(expr.name, true);
		
		if (idleft == null)
		{ throw new InterpretError("Left operand of assignment not found.");
		}

		if(DEBUG)
		{
			System.out.println("left operand of assignment is ");
			idleft.dump();
		}

		if (idleft.kind == STATE || (idleft instanceof DomInfo && ((DomInfo)idleft).dom.isStateDom == 1)) //second condition added by Yi Zheng May 14, 2002
		{
	//	if(DEBUG)
	//	 	System.out.println("In ApplyInOut(): idleft.kind is STATE");

		 //  	int stmtType = ((StateInfo) idleft).type; Yi Zheng May 14, 2002
			int stmtType;
			if (idleft.kind == STATE)
				stmtType = ((StateInfo) idleft).type; 
			else
				stmtType = ((DomInfo) idleft).dom.type;
		    
	  		NREnvironment NRenv = env.addNRFrame(rel); 
 		//	NREnvironment NRenv = tempEnv.addNRFrame(rel); 
			int nrow = rel.numtuples;
			
			for (NRenv.row = 0; NRenv.row < nrow; NRenv.row++)
			{
			  NRInfo nri = env.lookupNR(left.name, false); 
	//		NRInfo nri = tempEnv.lookupNR(left.name, false); 
			  col = nri.index;
			  NREnvironment nre = nri.NRenv;
			  
			  switch(stmtType)
			  { case INTEGER:
			  	//added by Yi Zheng April 18, 2002
			  	EvalExpr.WEIZHONG = true;
			 	 
			 	 ((int[]) nre.data[col])[nre.row] = EvalExpr.evalIntExpr(expr,env); 
	//		((int[]) nre.data[col])[nre.row] = EvalExpr.evalIntExpr(expr,tempEnv); 
			    
			  	//added by Yi Zheng April 18, 2002
			    EvalExpr.WEIZHONG = false;
			  
			  
			         break;
				 
			    default:
			    	throw new InterpretError("This type not	implemented yet.");
			  }	 
			}
			
			env.removeNRFrame(); 
//			tempEnv.removeNRFrame();

			rel.save();
//rel.printdata();		
		}
		else if (idleft.kind == PARAMETER)
		{ 
			int stmtType = ((ParamInfo) idleft).type;
			Environment e = ((ParamInfo) idleft).env;
			col = ((ParamInfo) idleft).index;

			switch(stmtType)
			{  case INTEGER:

				((int[]) e.data[col])[e.row] = EvalExpr.evalIntExpr(expr, env);
				break;

			   default:
			        throw new InterpretError("This type not	implemented yet.");
			}
			
			
		}
		else
		{ throw new InterpretError("Statement not supported yet."); 
		}
	     }
	     else 
	     { throw new InterpretError("Invalid assignment statement.");
	     }
	     	

		break;
	    default:
	        throw new InterpretError ("Statement type not implemented yet.");
	  }
	}  
        return rel; 
   }
*/

   /**
    * method called by applyInOut in updates
    * @author Yi Zheng June 10, 2002
    */
   private void runSingleStmtInUpdate(Relation rel, SimpleNode stmt, Environment tempEnv, NREnvironment NRe)
   throws InterpretError
   {
   	//addition by Yi Zheng May 20, 2002
    //remember the old interpreter.env
    Environment oldEnv = interpreter.env;
    //set interpreter.env to env
    //this is necessary due to the fact that some methods in Interpreter
    //uses the object variable env, instead of the one passed in as a parameter
    interpreter.env = tempEnv;
    //end of addition May 20, 2002
    int col;
    
    try{
		  switch(stmt.type)
		  { case OP_ASSIGNMENT :
		      if (stmt.jjtGetNumChildren() == 2) 
		      { 
		      	SimpleNode left = (SimpleNode) stmt.jjtGetChild(0);
				SimpleNode expr = (SimpleNode) stmt.jjtGetChild(1);
		
		 		IDInfo idleft = tempEnv.lookup(left.name, true);
				
				if (idleft == null)
				{ //must be on the fly declaration of relations on the LHS
				  //because scalars must be either declared as state or local
				  //scalars in parameter must be in parameters
					 Relation r=new Relation(left.name, new Domain[0], 0, 0, 0, tempEnv);
					 idleft = new RelInfo(r);
					 tempEnv.put(left.name, idleft);//will later be replaced
				}
		
				if (idleft.kind == NR || idleft.kind == STATE 
				   || (idleft instanceof DomInfo && ((DomInfo)idleft).dom.isStateDom == 1)) //second condition added by Yi Zheng May 14, 2002
				{
					int stmtType;
					if(idleft.kind == NR)
					{
						stmtType = ((NRInfo) idleft).type;
					}
					else if (idleft.kind == STATE)
					{
						stmtType = ((StateInfo) idleft).type; 
					}
					else
					{
	//					if(DEBUG)
	//						System.out.println(" HIDDEN STATE DOM");
						stmtType = ((DomInfo) idleft).dom.type;
					}
				    
				  //if a computation is making changes to a state from inside an update
				  //the state must have been exported to a relation already
				  //so it must have been an NRInfo
				  NRInfo nri = null;
				  if (idleft.kind == NR)
				  	nri = (NRInfo) idleft;
				  else
				  	nri = tempEnv.lookupNR(left.name, false); 
				  if(nri == null)
				  	throw new InterpretError(left.name + " not found in " + rel.name);
				  col = nri.index;
				  NREnvironment nre = nri.NRenv;
	
				  //added by Yi Zheng April 18, 2002
				  EvalExpr.WEIZHONG = true;
				  
				  switch(stmtType)
				  { case INTEGER:
						((int[]) nre.data[col])[nre.row] = EvalExpr.evalIntExpr(expr,tempEnv); //Yi Zheng, June 2, 2002
				         break;
					 //except for INTEGER, all other types are added by Yi Zheng June 2, 2002
					 case BOOLEAN:
					 	throw new InterpretError("To be implemented");
					 case SHORT:
					 	((short[]) nre.data[col])[nre.row] = EvalExpr.evalShortExpr(expr, tempEnv);
					 	break;
					 case LONG:
					 	((long[]) nre.data[col])[nre.row] = EvalExpr.evalLongExpr(expr, tempEnv);
					 	break;
					 case FLOAT:
					 	((float[]) nre.data[col])[nre.row] = EvalExpr.evalFloatExpr(expr, tempEnv);
					 	break;
					 case DOUBLE:
					 	((double[]) nre.data[col])[nre.row] = EvalExpr.evalDoubleExpr(expr, tempEnv);
					 	break;
				//	 case UNIVERSAL:
				//	 case ATTRIBUTE:
					 case STRING:
					 	((String[]) nre.data[col])[nre.row] = EvalExpr.evalStringExpr(expr, tempEnv);
					 	break;
					 case IDLIST:
						Relation nestrtmp = null;
						Relation nestrel = null;
					    String ttmpname1 = interpreter.nextTempName();
					    String ttmpname2 = interpreter.nextTempName();
					    String ttmpname3 = interpreter.nextTempName();
					    String ttmpname4 = interpreter.nextTempName();
					    long surrogate, newSur = 0;
					    Domain[] iiddom = new Domain[1];
	                    iiddom[0] = tempEnv.lookupDom(".id", true);
					   
					   // extract the nested relation
					   if ((nestrel = tempEnv.lookupNestedRel(left.name, true)) == null) 		
					   {	
						throw new InterpretError("Nested Relation '"+left.name+"' does not exist");
					   } 
	
					   if ((surrogate = tempEnv.getSurrogate(left.name)) == -1) 
					   {
						throw new InterpretError("Surrogate not found.");
					   }
					   
					   if(surrogate == 0)
					   {
					   	newSur = Surrogate.next();
					   	tempEnv.setSurrogate(left.name, newSur);
					   }
					   
					   if((nestrtmp = tempEnv.lookupRelSurr(left.name, true)) == null) //nestrtmp.name == ""
					   {  throw new InterpretError(
							  "Relation '"+left.name+"' does not exist");//changed from id.name by Yi Zheng June 2, 2002
					   }
					   
					   if(nestrtmp.numtuples == 0)
					   {
					   	Object[] dataArray = nestrtmp.createDataArray(nestrtmp.domains, 1);
					   	for(int k = 0; k < nestrtmp.numattrs; k++)
					   	{
					   		if(nestrtmp.domains[k].name.compareTo(".id") == 0)
					   		{
					   			((long[]) dataArray[k])[0] = (surrogate == 0? newSur : surrogate);
					   			break;
					   		}
					   	}
					   	
					   	if(nestrtmp.name.compareTo("") == 0)
					   		nestrtmp.name = interpreter.nextTempName();
					   	
					   	nestrtmp.append(dataArray, 1);
					   	if(nestrtmp.data == null) nestrtmp.load();//should never be here,as append always leaves data in
					   	nestrtmp.save();//save also leaves data not null
					   }
					   else
					   {   	if(nestrtmp.name.compareTo("") == 0)
					   		nestrtmp.name = interpreter.nextTempName();
					   		nestrtmp.save();
					   }

					   if(nestrel.data == null) nestrel.load();
					   nestrel.save();
					   if(nestrtmp.data == null) nestrtmp.load();
					   nestrtmp.save();
					   
					   Relation rrest = nestrel.mujoin(OP_DJOIN, null, nestrtmp, null, ttmpname4);
	
						if(rrest.name.compareTo("") == 0)
							rrest.name = ttmpname4;
						rrest.load(true);
						rrest.save();	
						
						//addition by Yi Zheng May 17, 2002, to transform the tree
						//such that all occurrences of IDENTIFIER are checked for possible
						//replacement by constants found in various IDInfo objects
						transformExpr(expr, tempEnv);
						//end of addition May 17, 2002
	
					   
					   Relation rtmp_tmp1 = interpreter.evaluateTLExpression(expr, ttmpname1, new Hashtable(), tempEnv); //no name
						if(rtmp_tmp1.name.compareTo("") == 0)
							rtmp_tmp1.name = ttmpname1;
						
					   // check whether nestrtmp and rtmp_tmp1 is the same type relation (attributes are of same type)	
					   // if not, throw an error
					   if (Interpreter.CompareRelTypeComp(nestrtmp, rtmp_tmp1) != 1) 
					     throw new InterpretError("The assignment is not allowed since the two operand relations are not of the same type!");
					   
	                   // project nestrtmp on .id
					   String idtemp = interpreter.nextTempName();
					   Relation idrel = nestrtmp.project(iiddom, idtemp);
					   if(idrel.name.compareTo("") == 0)
					   	idrel.name = idtemp;
					   if(idrel.data == null) idrel.load();
					
					   if(rtmp_tmp1.data == null)
							rtmp_tmp1.load();
					   
					   // (add domain .id to rtmp_tmp1) -> rtmp_tmp2	
					   Relation rtmp_tmp2  = idrel.mujoin(OP_IJOIN, null, rtmp_tmp1, null, ttmpname2);
						if(rtmp_tmp2.name.compareTo("") == 0)
							rtmp_tmp2.name = ttmpname2;
						rtmp_tmp2.load(true);
						rtmp_tmp2.save();
						
					   //deal with <+
					   if(rtmp_tmp2.data == null)
					   		rtmp_tmp2.load();
					   if(nestrtmp.data == null)
					   		nestrtmp.load();
					   if(stmt.opcode == OP_ASSIGNADD)
					   	nestrtmp.assignAdd(null, rtmp_tmp2, null);   // which is left side?
					   
					   if (nestrtmp.name == "")
					     nestrtmp.name = "Temp";
					   String nestrtmp_name = nestrtmp.name;
			   		   tempEnv.removeRel(nestrtmp_name, false);
			   		   
			   		   // rename back to nestrtmp relation
			   		   tempEnv.put(rtmp_tmp2.assign(null, nestrtmp_name, null));
					   nestrtmp = tempEnv.lookupRel(nestrtmp_name, true);
					   nestrtmp.load(true);
					   nestrtmp.save();
	
					   if(rrest.data == null)
					   	rrest.load();
					   
					   String ttmpname5 = interpreter.nextTempName();
					   Relation temp_nestrel = rrest.mujoin(OP_UJOIN, null, nestrtmp, null, ttmpname5);
					   temp_nestrel.load(true);
					   nestrel.appendInRAM(temp_nestrel, nestrel);		
					   nestrel.load(true);
					   nestrel.save();
					   
					   interpreter.deleteFile(ttmpname1);
					   interpreter.deleteFile(ttmpname2);
					   interpreter.deleteFile(ttmpname3);
					   interpreter.deleteFile(ttmpname4);
					   interpreter.deleteFile(ttmpname5);
					   break;
					 //end of addition by Yi Zheng June 2, 2002
				    default:
				    	throw new InterpretError("This type not	implemented yet.");
				  }	 
				  //added by Yi Zheng April 18, 2002
				  EvalExpr.WEIZHONG = false;
				}
				else if (idleft.kind == PARAMETER) //not used by applyInOut, Yi Zheng June 2, 2002
				{ throw new InterpretError("unexpected use of Parameter info in runSingleStmtInUpdate");
				}
				else if (idleft.kind == LOCAL) //added by Yi Zheng June 2, 2002
				{
					int stmtType = ((LocalInfo) idleft).type;
					
					switch(stmtType)
					{
						case INTEGER:
							((LocalInfo) idleft).intVal = EvalExpr.evalIntExpr(expr, tempEnv);
							break;
							
						//addition by Yi Zheng June 2, 2002
						 case BOOLEAN:
						 	throw new InterpretError("To be implemented");
						 case SHORT:
						    ((LocalInfo) idleft).intVal = EvalExpr.evalShortExpr(expr, tempEnv);
						    break;
						 case LONG:
						    ((LocalInfo) idleft).longVal = EvalExpr.evalLongExpr(expr,tempEnv);
							break;
						 case FLOAT:
						    ((LocalInfo) idleft).floatVal = EvalExpr.evalFloatExpr(expr,tempEnv);
							break;
						 case DOUBLE:
						    ((LocalInfo) idleft).doubleVal = EvalExpr.evalDoubleExpr(expr,tempEnv);
							break;
					//	 case UNIVERSAL:
					//	 case ATTRIBUTE:
						 case STRING:
		 				    ((LocalInfo) idleft).stringVal = EvalExpr.evalStringExpr(expr,tempEnv);
							break;
						 case IDLIST:
						 	transformExpr(expr, tempEnv);
						 	String tname = interpreter.nextTempName();
						    Relation rightrel = interpreter.evaluateTLExpression(expr, tname, new Hashtable(), tempEnv); 
		
						  	//there must be a relation in the env that was put in by processCallParams
						  	//so find it and assign it relResult
						  	Relation leftrel = tempEnv.lookupRel(left.name, true);
						  	if (leftrel == null)
						  		throw new InterpretError(left.name + " as a relation not found");
		
						 	//addition by Yi Zheng May 20, 2002
						 	if(stmt.opcode == OP_ASSIGNADD)
                                                        // bug fix by Zongyan Nov 29, 2004
						  		rightrel.assignAdd(null, leftrel, null);
						  	//	leftrel.assignAdd(null, rightrel, null);
						 	//end of addition May 20, 2002

						  	//leftrel = rightrel.duplicate(true);  
						  	//leftrel.name = left.name;
						  	//if (leftrel.data == null) leftrel.load();
						  	//leftrel.save();
						  	//((LocalInfo)idleft).relVal = leftrel;
						  
                                                        Relation r;
						 	if(stmt.opcode == OP_ASSIGNADD)
						  	    r = leftrel.duplicate(true);  
                                                        else
						  	    r = rightrel.duplicate(true);  
						  	r.name = left.name;
						  	if (r.data == null) r.load();
						  	r.save();
						  	((LocalInfo)idleft).relVal = r;
						        // end of bug fix Nov 29, 2004 

						  	break;
		 
						 //end of addition by Yi Zheng June 2, 2002
		
						default:
							throw new InterpretError("not implemented yet");
					}
				} 
				else if(idleft.kind == REL) //declared on the fly in the block
				{
				 	transformExpr(expr, tempEnv);
				 	String tname = interpreter.nextTempName();
				    Relation rightrel = interpreter.evaluateTLExpression(expr, tname, new Hashtable(), tempEnv); 
				
					//addition by Yi Zheng May 20, 2002
					Relation leftrel = tempEnv.lookupRel(left.name, true);
				  	if (leftrel == null)
				  		throw new InterpretError(left.name + " as a relation not found");
		
					 if(stmt.opcode == OP_ASSIGNADD)
					 //bug fix by Zongyan Nov 29 , 2004			
					 //leftrel.assignAdd(null, rightrel, null);
					   rightrel.assignAdd(null, leftrel, null);
					 //end of addition May 20, 2002
		
		
				 	//Relation r = rightrel.duplicate(true); // bug fix by Zongyan Nov 29, 2004
                                        Relation r;
					if(stmt.opcode == OP_ASSIGNADD)
				 	     r = leftrel.duplicate(true);
                                        else
				 	     r = rightrel.duplicate(true);
				   	r.name = left.name;
					if (r.data == null) r.load();
				   	((RelInfo) idleft).rel = r;
				   	tempEnv.put(left.name, idleft); //overwrites the initial entry
				        r.save();
				}
				else
				{ throw new InterpretError("Statement not supported yet."); 
				}
			     }
			     else if(stmt.jjtGetNumChildren() == 4)
			     {
			     	throw new InterpretError("To be implemented");
			     }
			     else 
			     { throw new InterpretError("Invalid assignment statement.");
			     }
			     	
				break;
			case OP_COMMAND:{ //added by Yi Zheng June 10, 2002
				switch(stmt.opcode)
				{
					case OP_PR:{
						SimpleNode firstchild = (SimpleNode) stmt.jjtGetChild(0);
						transformExpr(firstchild, tempEnv);
						interpreter.executeCompCommand(stmt, tempEnv);
						break;
					}
					case OP_INPUT:
					case OP_DELD:
					case OP_DELR:
						throw new InterpretError("Command type not implemented.");
					default:
						interpreter.executeCompCommand(stmt, tempEnv);
						break;
				}
				break;
			}
			case OP_DECLARATION:{
				switch(stmt.opcode)
			   	{  
			   		case OP_COMPUTATION :
					 String compName = ((SimpleNode) stmt.jjtGetChild(0)).name;
		                         // weizhong
					 SimpleNode tempNode =(SimpleNode)stmt.jjtGetChild(0);
					 while (tempNode.name == null) {
					   tempNode = (SimpleNode) tempNode.jjtGetChild(0);
					 }
					 compName = tempNode.name;
					 if(NRe.row == 0)
					 {  //This is our first encounter with this declaration
						//so we must process it.
						Computation c = new Computation(stmt, tempEnv);
						tempEnv.put(compName, new RelInfo(c));
						
						//Since this is inside update, we assume no computation is exported
					 }
					 break;
				case OP_VIEW :{
					SimpleNode init_node = (stmt.isBitSet(0)?(SimpleNode)stmt.jjtGetChild(1):null);
					SimpleNode target_node = null;
					SimpleNode is_node = null;
					
					if(init_node != null)
					{
						is_node = (SimpleNode)stmt.jjtGetChild(2);
						target_node = (stmt.isBitSet(1)?(SimpleNode)stmt.jjtGetChild(3):null);
					}else
					{
						is_node = (SimpleNode)stmt.jjtGetChild(1);
						target_node = (stmt.isBitSet(1)?(SimpleNode)stmt.jjtGetChild(2):null);
					}
					
					transformExpr(is_node, tempEnv);
					if(init_node != null) transformExpr(is_node, tempEnv);
					if(target_node != null) transformExpr(target_node, tempEnv);
					interpreter.executeCompDeclaration(stmt, tempEnv);
					break;
				}
				case OP_LET :{
					SimpleNode be_node = null;
					SimpleNode init_node = (stmt.isBitSet(0)?(SimpleNode)stmt.jjtGetChild(1):null);
					if(init_node != null)
						be_node = (SimpleNode) stmt.jjtGetChild(2);
					else
						be_node = (SimpleNode) stmt.jjtGetChild(1);
					transformExpr(be_node, tempEnv);
					if(init_node != null) transformExpr(init_node, tempEnv);
					interpreter.executeCompDeclaration(stmt, tempEnv);
					break;
				}
				  case OP_RELATION :
				  case OP_DOMAIN :
					//addition by Yi Zheng May 20, 2002
				  	 interpreter.executeCompDeclaration(stmt, tempEnv);
				  	 break;
				  	//end of addition May 20, 2002
				  default :
					 throw new 
					 InterpretError("Not implemented yet");
			   }
			   break;
			}
			case OP_STATEMENT:{
			   switch(stmt.opcode) 
			   {  case OP_CONDITIONAL :
					 SimpleNode ifNode, thenNode, elseNode;
					 ifNode = (SimpleNode) stmt.jjtGetChild(0);
					 thenNode = (SimpleNode) stmt.jjtGetChild(1);
			 
					 if (EvalExpr.evalBooleanExpr(ifNode, tempEnv))
					 {  runSingleStmtInUpdate(rel, thenNode, tempEnv, NRe);
					 }
					 else if(stmt.jjtGetNumChildren() == 3)
					 {  elseNode = (SimpleNode) stmt.jjtGetChild(2);
						runSingleStmtInUpdate(rel, elseNode, tempEnv, NRe);
					 }
					 break;
				  case OP_STMTBLOCK :
					 for(int stmtNo=0; stmtNo<stmt.jjtGetNumChildren(); stmtNo++)
					 {  runSingleStmtInUpdate(rel, (SimpleNode) stmt.jjtGetChild(stmtNo), tempEnv, NRe);
					 }
					 break;
				  case OP_COMPCALL: //added by Yi Zheng May 20, 2002
				  	interpreter.executeCompStatement(stmt, tempEnv);
				  	break;
				  case OP_SEQUENCE :
				  case OP_PARALLEL :
				  case OP_CHOICE :
				  case OP_FORLOOP :
				  case OP_WHILELOOP :
				  case OP_EXIT :
				  case OP_DEADLOCK :
				  case OP_EXEC :
					 throw new 
					 InterpretError("Not implemented yet");
			   }
							
				break;
			}
			case OP_UPDATE:
			   //added by Yi Zheng May 20, 2002
				Environment uenv = new Environment(tempEnv, tempEnv.getTable(), tempEnv.reltable, tempEnv.domtable);

				switch(stmt.opcode)
				{
					case OP_CHANGE:{
						SimpleNode stmtList = (stmt.isBitSet(0)?(SimpleNode)stmt.jjtGetChild(1):null);
						SimpleNode using = null;
						if(stmtList != null && stmt.isBitSet(1))
							using = (SimpleNode)stmt.jjtGetChild(2);
						else if(stmt.isBitSet(1))
							using = (SimpleNode)stmt.jjtGetChild(1); //impossible
						if(stmtList != null)
						{
							for(int i = 0; i < stmtList.jjtGetNumChildren(); i++)
								transformStmt((SimpleNode)stmtList.jjtGetChild(i), tempEnv);
						}
						if(using != null)
						{
							int numChild = using.jjtGetNumChildren();
							if(numChild == 1)
							{
								transformExpr((SimpleNode)using.jjtGetChild(0), tempEnv);
							}
							else if(numChild == 3)
							{
								transformExpr((SimpleNode)using.jjtGetChild(2), tempEnv);
							}
							else throw new InterpretError("invalid update change");
						}
						break;
					}
					case OP_ADD:
					case OP_DELETE:{
						int numChild = stmt.jjtGetNumChildren();
						if(numChild == 2)
						{
							transformExpr((SimpleNode)stmt.jjtGetChild(1), tempEnv);
						}
						else if(numChild == 4)
						{
							transformExpr((SimpleNode)stmt.jjtGetChild(3), tempEnv);
						}
						else throw new InterpretError("invalid update add/delete");
						break;
					}
					default:
						throw new InterpretError("invalid update type");
				}

				// produce trigger tree
				int diftrig = interpreter.lookCompUpdate(stmt, uenv, null, uenv.triggerroot);
				// reverse the result -- undo trigger
				interpreter.doCompTrigger(tempEnv, uenv.triggerroot, 1);
				if (diftrig == 1)
				  interpreter.executeCompUpdate(stmt, stmt, uenv); 
				interpreter.RemoveTrigger(uenv.triggerroot);		// clear up
				break;
				  //end of addition May 20, 2002

			//end of Yi's addition June 10, 2002
		    default:
		        throw new InterpretError ("Statement type not implemented yet.");
		  }
   	}
   	catch(InterpretError inte)
   	{
   		interpreter.env = oldEnv;
   		throw inte;
   	}
   	interpreter.env = oldEnv;
   	
   } /* runSingleStmtInUpdate */
   
   /**
	* Check if this computation is being run in constraint mode
	* @param   joinAttrs   Array of input domains.
	
   private boolean constraintMode(Domain[] joinAttrs, Environment env)
   throws InterpretError
   {  int index;
	  int type = 0;
	  IDInfo id;
	  ParamInfo pi;
	  for(int i=0; i<joinAttrs.length; i++)
	  {  pi = env.lookupParam(joinAttrs[i].name, false);
		 if(pi == null) //should never happen
		 {  throw new InterpretError("Invalid join input attribute");
		 }
		 index = pi.index;
		 type |= 1<<index;
	  }
	  for(int i=0; i<blocks.length; i++)
	  {  if (type == blocks[i].blockType)
		 {  return false;
		 }
		 else if ((type & blocks[i].blockType) == blocks[i].blockType)
		 {  return true;
		 }
	  }
	  return false; //This case should never happen
   }   */ //code by Baker
   
   /**
	* Check if this computation is being run in constraint mode
	* @param   joinAttrs   Array of input domains. 
	* @author Yi Zheng March 31, 2002
	*/
   private boolean constraintMode(Domain[] joinAttrs, Environment env)
   	{ return env.constraint_mode;}
   
   
   /**
	* Return the domain array whose elements are those which are in domb
	* but not in doma.
	* @author Patrick Baker
	*/
   private Domain[] extraDomains(Domain[] doma, Domain[] domb)
   {  Domain[] dom = new Domain[domb.length]; 
	  int j=0; 
	  for (int i=0; i<domb.length; i++) 
	  {  if (!isDomainIn(domb[i], doma))
		 {  dom[j++] = domb[i];
		 }
	  }
	  Domain[] tmp = new Domain[j];
	  for (int i=0; i<j; i++)
	  {  tmp[i] = dom[i];
	  }
	  return tmp;
   }   /* extraDomains */
   
   /**
    * used by applyIjoin where there are named join attrs
    * @author Yi Zheng June 20, 2002
    */
   private Domain[] extraDomains2(Domain[] doma, Domain[] domb, Hashtable tbl)
   {
   	Domain[] dom = new Domain[domb.length];
   	int j = 0;
   	for (int i = 0; i < domb.length; i++)
   	{
   		Domain lookupD = (Domain) tbl.get(domb[i].name);
   		if (lookupD != null) //domb[i] is a joinAttr, so not extra
   			continue;
   		if(!isDomainIn(domb[i], doma))
   			dom[j++] = domb[i];
   	}
   	Domain[] tmp = new Domain[j];
   	for(int i = 0; i < j; i++)
   		tmp[i] = dom[i];
   	return tmp;
   } /* extraDomains2 */
   
   /**
	* Decide which "alt" block to use this time around.
	* @param   joinAttrs   Array of input domains.
	* @author Patrick Baker
	* modified by Yi Zheng
	*/
   private CompBlock findBlock(Domain[] joinAttrs, Environment env)
   throws InterpretError
   {  int index;
	  int type = 0;
	  IDInfo id;
	  ParamInfo pi;
	  for(int i=0; i<joinAttrs.length; i++)
	  {  pi = env.lookupParam(joinAttrs[i].name, false);
		 if(pi == null) //should never happen
		 {  throw new InterpretError("Invalid join input attribute");
		 }
		 index = pi.index;
		 type |= 1<<index;
	  }
	  
	  
	  //added by Yi Zheng March 31, 2002
	  for(int i=0; i<blocks.length; i++)
	  {  if (type == blocks[i].blockType) //exact match, including the case where all params are inputs (all 1's) or outputs (0)
		 {  env.constraint_mode = false;
		 	return blocks[i];
		 }
	  }
	  
	  //assuming that no exact block type is found, try constraint mode
	  ///since the first for loop makes sure that every attribute in joinAttrs is a legal parameter
	  //we can think of the case where all parameters are provided as the constraint mode
	  //we further assume that when this is the case we can pick the first block to 
	  //check for validity of the inputs (whether they satisfy the constraint, leaving
	  //the consistency of definitions of the blocks to the programmer
	  if(joinAttrs.length == numattrs)
	  {
	  	env.constraint_mode = true;
	  	return blocks[0];
	  }
	  //end Yi's addition
/*
	  for(int i=0; i<blocks.length; i++)
	  {  if ((type & blocks[i].blockType) == blocks[i].blockType)
		 { 
		  return blocks[i];
		 }
	  }
	  */ //commented out by Yi Zheng March 31, 2002
	  throw new InterpretError("No matching block type in computation");
   }   /* findBlock */
   
   /**
	* Initialize the data array. This is called only when
	* this computation is invoked by an ijoin with some relation.
	* @param   rel   The relation that is ijoined with this computation.
	* @author Patrick Baker
	* modified by Yi Zheng
	*/
   private void initializeData(Relation rel, Environment env, Domain[] tempJoinAttrs)//June 21, 2002
 //  private void initializeData(Relation rel, Environment env, Hashtable tbl)
   throws InterpretError
   {  //make a copy of the input relation's data
//	  Object[] tempdata = rel.duplicate(true).data; //modified by Yi Zheng June 18, 2002
	 String tmpRelName = Interpreter.nextTempName();
	 Relation tmpRel = rel.project(rel.domains, tmpRelName);
	 if(tmpRel.data == null)
		tmpRel.load();
	
	 Object[] tempdata = tmpRel.data;

	  numtuples = rel.numtuples;
	  		
	  //We must also make columns for state variables
	  String[] states = env.getStateNamesLocal();
	  int numStates = states.length;
 	  data = new Object[numattrs + extraAttrs.length + numStates];
	  Domain[] newDomains = new Domain[numattrs+extraAttrs.length+numStates];

	  //sort the columns so that they are in the order specified by the 
	  //computation declaration. Extra fields are tacked on at the end
	  int i;
	  for(i=0; i<numattrs; i++)
	  {  Domain d = domains[i];
		 newDomains[i] = d;
		 
	//	 if(isDomainIn(d,joinAttrs)) //Yi Zheng June 21, 2002
		 if(isDomainIn(d, tempJoinAttrs))
		 {  //REVISIT: watch for alias
			//This is an input domain, 
			//we must find the corresponding column in tempdata
			int j;
			
			for(j=0; !(rel.domains[j].name.compareTo(d.name) == 0); j++)
			{}
			data[i] = tempdata[j];
		 }
		 else
		 {  //This is an output domain, we must create an array for it
			switch(d.type)
			{  case INTEGER :
			   {  data[i] = new int[numtuples];
				  break;
			   }
			   case ATTRIBUTE://treat as String, Yi Zheng May 24, 2002
			   case UNIVERSAL://treat as String, Yi Zheng May 24, 2002
			   case STRING :
			   case TEXT :
			   case STMT :
			   case EXPR : 
			   {  data[i] = new String[numtuples];
				  break;
			   }
			   case BOOLEAN : 
			   {  data[i] = new byte[numtuples];
				  break;
			   }
			   case SHORT : 
			   {  data[i] = new short[numtuples];
				  break;
			   }          
			   case LONG :
			   case IDLIST :
			   case COMP :
			   {  data[i] = new long[numtuples];
				  break;
			   }
			   case FLOAT : 
			   {  data[i] = new float[numtuples];
				  break;
			   }
			   case DOUBLE : 
			   {  data[i] = new double[numtuples];
				  break;
			   }
			} /*switch */
		 }
	  }
	  //Now tack on the extra domains
	  int k=0;
	  for(; i<numattrs + extraAttrs.length;i++)
	  {  Domain d = extraAttrs[k];
		 newDomains[i] = d;
		 k++;
		 int j;
		 for(j=0; !(rel.domains[j].name.compareTo(d.name) == 0); j++)
		 {}
		 data[i] = tempdata[j];
	  }
	  //Finally, tack on domains for state vars
	  //Note that state vars were added to callEnv as domains in initializeEnvironment
	  //when the computation is first declared
	  k=0;
	  for(; i<data.length;i++)
	  {  Domain d = env.lookupDom(states[k], true);
	  
	  //added by Yi Zheng April 5, 2002
	  	if(d == null) 
	  		throw new InterpretError("Serious error: state var '" + states[k] + "' not found as a domain");
	  	StateInfo sinf = (StateInfo) env.lookup(states[k], false);
	  	sinf.index = i;
	  	
		 newDomains[i] = d;
		 k++;
		 switch(d.type)
		 {  case INTEGER :
			{  data[i] = new int[numtuples];
			   break;
			}
			case ATTRIBUTE://treat as STRING, Yi Zheng May 24, 2002
			case UNIVERSAL://treat as STRING, Yi Zheng May 24, 2002
			case STRING :
			case TEXT :
			case STMT :
			case EXPR : 
			{  data[i] = new String[numtuples];
			   break;
			}
			case BOOLEAN : 
			{  data[i] = new byte[numtuples];
			   break;
			}
			case SHORT : 
			{  data[i] = new short[numtuples];
			   break;
			}
			case LONG :
			case IDLIST :
			case COMP :
			{  data[i] = new long[numtuples];
			   break;
			}
			case FLOAT : 
			{  data[i] = new float[numtuples];
			   break;
			}
			case DOUBLE : 
			{  data[i] = new double[numtuples];
			   break;
			}
		 } /*switch */
	  }

	  numattrs += extraAttrs.length+numStates;
	  //Note that here we are changing domains and numattrs.
	  //These will have to be restored when we finish running 
	  //this computation, but before returning the result to the 
	  //interpreter. Check out the applyIjoin and applySelect methods.
	  domains = newDomains; 
   }   /* initializeData */
   
   /**
    * called by apply* methods
    * @author Patrick Baker
    * modified by Yi Zheng
    */
   private Environment initializeEnvironment(SimpleNode node,
									   Environment env, Environment callEnv)
   throws InterpretError
   {  int paramIndex = 0;
	  //Dissect the parts of the computation from the syntax tree
	  SimpleNode identifier;  //name of the computation
	  SimpleNode parameters;  //note: The parameter list could be empty
	  SimpleNode body;        //the declaration section and 'alt' blocks
	  if (node.jjtGetNumChildren() == 3)
	  {  body       = (SimpleNode) node.jjtGetChild(2);
		 parameters = (SimpleNode) node.jjtGetChild(1);
		 identifier = (SimpleNode) node.jjtGetChild(0);
	  }
	  else if (node.jjtGetNumChildren() == 2)
	  {  //Empty parameter list in this case
		 body       = (SimpleNode) node.jjtGetChild(1);
		 parameters = null;
		 identifier = (SimpleNode) node.jjtGetChild(0);
	  }
	  else
	  {  //The Parser should really prevent this from ever happening
		 throw new 
		 InterpretError("Error: Incorrect computation definition");
	  }

	  //Now fill in the environment
	  //Process the parameters
	  if (parameters != null)
	  { 
		 int numChildren = parameters.jjtGetNumChildren();
		
		 for(paramIndex=0; paramIndex<numChildren; paramIndex++)
		 {  String paramName = 
					 ((SimpleNode) parameters.jjtGetChild(paramIndex)).name;
			Domain d = domains[paramIndex];
			ParamInfo pi = new ParamInfo(d.type, paramIndex);
			env.put(paramName, pi);

			//Optimization: Create relation objects for the parameters
			//and store these in the ParamInfo objects. This way, we
			//won't have to create new Relation objects each time we call
			//extract().
			//REVISIT: only need to do this for input params, if possible

/*
//REALLY should get rid of this optimization! Yi Zheng March 25, 2002

			if (d.type == IDLIST)
			{  Relation dotRel = env.lookupRel("."+paramName, true);
//REVISIT: get rid of this optimization
if(dotRel != null)
{

			   Relation r = dotRel.duplicate(false);//don't load data
			   r.name = "";	 //REVISIT:why is the name empty? Yi Zheng

			   //remove the ".id" field from r
			   r.numattrs--;
			   Domain[] doms = new Domain[r.numattrs];
			   int j=0;
			   for (int k=0; k<dotRel.numattrs; k++)
			   {
			   	

			   	if(DEBUG)
			   	{
			   		System.out.println("dotRel = " + dotRel.name);
			   		if(r.domains == null)
			   		System.out.println("r.domains is null");
			   		else if(r.domains[k] == null)
			   		System.out.println("r.domains[" + k + "] is null");
			   		else if(r.domains[k].name == null)
			   		System.out.println("r.domains[" + k + "].name is null");
			   	}

			   	  if (!r.domains[k].name.equals(".id"))
				  {  doms[j++] = r.domains[k];
				  }
			   }
			   r.domains = doms;
			   r.createDataArray(r.numtuples);
			   pi.rel = r;
}

			}
*/
		 }
	  }
	  
	  //Process local and state variable declarations
	  if (body.jjtGetNumChildren() <= 0)
	  {  //The Parser should really prevent this from ever happening
		 throw new
		 InterpretError("Incorrect computation definition\n" +
						"Computation has no body");
	  }
	  SimpleNode decl = (SimpleNode) body.jjtGetChild(0);
	  if( decl.type == OP_COMPVARDECL)
	  {  //The optional declaration section is in fact present
		 for(int i=0; i<decl.jjtGetNumChildren();i++)
		 {  SimpleNode vardecl = (SimpleNode) decl.jjtGetChild(i);
			if(vardecl.type == OP_LOCALVARDECL ||
			   vardecl.type == OP_STATEVARDECL)
			{  //Found a local or state variable declaration
			   SimpleNode IDList;
			   
			   //added by Yi Zheng May 15, 2002, for handling states of IDLIST type
				Vector domlist = new Vector();
		 		Domain[] domarry = null;
				SimpleNode idList;
				//end of addition May 15, 2002
				
			   int type = ((SimpleNode) vardecl.jjtGetChild(1)).opcode;
			   //Check if declarations are of allowed type
			   switch(type)
			   {  case OP_COMP :
				   throw new
				   InterpretError("Cannot declare a computation in the declaration section");
				  case OP_IDLIST : {
					//added by Yi Zheng May 15, 2002, for handling states of type IDLIST
				
					//get the domains the state or local var is defined on
					SimpleNode typeNode = (SimpleNode) vardecl.jjtGetChild(1);
					if(typeNode.jjtGetNumChildren() < 1)
						throw new InterpretError("Empty domain list in IDLIST typed state declaration");
					idList = (SimpleNode) typeNode.jjtGetChild(0);			
					for(int ii = 0; ii < idList.jjtGetNumChildren(); ii++)
	     			{  String dname = ((SimpleNode) idList.jjtGetChild(ii)).name;
	     
	     			// Check validation of the dom name. But note, we can't
	     			// increment the domain's numref at the moment; We have
	     			// to complete _all_ validation check. If the check
	     			// fails, an exception is thrown and nothing happens.
	     				Domain oldD = env.lookupDom(dname, true);
	    				if(oldD == null)
	       				{  throw new 
		    				InterpretError("domain "+ dname + " has not been declared");
	       				}
	     				else
	       				{  
	       					domlist.addElement(oldD);
	       				}
	     			}
	   				domarry= new Domain[domlist.size()];
	   				domlist.copyInto(domarry);
	   				
					break;
					//end of addition May 15, 2002
					}
				  case OP_TEXT :
				  case OP_STMT :
				  case OP_EXPR :
				   throw new
				   InterpretError("Types 'text', 'stmt', and 'expr' are not yet " +
								  "supported in computations");
			   }
 
			   //Now add the varibles to the environment
			   IDList = (SimpleNode) vardecl.jjtGetChild(0);
			   if(vardecl.type == OP_LOCALVARDECL)
			   {
			   	for(int j=0; j<IDList.jjtGetNumChildren(); j++)
			   	{  SimpleNode id = (SimpleNode) IDList.jjtGetChild(j);
				  LocalInfo li = new LocalInfo(type-OP_BOOLEAN, null);
				  //added by Yi Zheng May 15, 2002
				  if (type == OP_IDLIST)
				  {
				  	li.relVal = new Relation(id.name, domarry, RELATION,
						0, domarry.length, 0, null, env);
					env.put(li.relVal);
				  }
				  else //assuming it is not OP_COMP
				  {
				  	Domain localdomain = new Domain(id.name, type-OP_BOOLEAN, 0, null);
				  	env.put(localdomain);
				  }
				  //end of addition May 15, 2002
				  env.put(id.name, li);
			   	}
			   }
			   else //must be a state variable declaration
			   for(int j=0; j<IDList.jjtGetNumChildren(); j++)
			   {  SimpleNode id = (SimpleNode) IDList.jjtGetChild(j);
				  String stateName = id.name;
				  
				  //it seems like the state vars are listed as parameters
				  //or treated like parameters that can be passed up to parent
				
					//added by Yi Zheng April 18, 2002 for STATE bug fix
					IDInfo oldSi = stateEnv.lookup(stateName,false);
					if(oldSi == null || oldSi.kind != STATE)
					{//end of addition by Yi April 18
					
					  StateInfo si = new StateInfo(type-OP_BOOLEAN, paramIndex);
					  paramIndex++;
					  
					  //added by Yi Zheng May 15, 2002
					  if (type == OP_IDLIST && (!UPDATE))
					  {
					  	si.relVal = new Relation(stateName, domarry, RELATION,
							0, domarry.length, 0, null, env);
					  	env.put(si.relVal);
					  	
					  }
					  
					  else if(type != OP_IDLIST) //assuming it's not OP_COMP
					  {
					  	Domain statedomain = new Domain(stateName,type,0,null);
					  	env.put(statedomain);
					  }
					  //end of addition May 15, 2002
					  env.put(stateName, si);
					  
					  stateEnv.put(stateName, si);//addition by Yi April 18
					}
					else 
					{
						if((!UPDATE) && (type == OP_IDLIST))
							env.put(((StateInfo)oldSi).relVal);
						env.put(stateName, (StateInfo)oldSi);
					
						
					}//end of addition by Yi April 18, 2002
				
				  //Need to create the hidden domain  in the parent
				  //environment if it doesn't already exist
				  if(callEnv != null && //short-circuit boolean operator
					 callEnv.lookupDom(stateName, true) == null) //CONTINUE MAY 30
				  { 
				  //if... then switch added by Yi Zheng April 23, 2002
				  //isStateDom flag added by Yi Zheng April 23, 2002
				  if(type == OP_IDLIST)
				  {//note we need to invoke a different Domain constructor
					//added by Yi Zheng May 15, 2002
					Domain d = new Domain(stateName, IDLIST, domarry, callEnv);
					d.isStateDom = 1;
					callEnv.put(d);
					//end of addition May 15, 2002
				  }
				  else
				  {
				   Domain d = new Domain(stateName,type-OP_BOOLEAN,0,null);//need to set a flag indicating that this is a state var with value!!! April 18, 2002
					d.isStateDom = 1;
					 callEnv.put(d);
				  }
				}
			   }
			}
			else
			{ //The Parser should really prevent this from ever happening 
			  throw new
			  InterpretError("Error: Incorrect computation definition\n"
						   + "Unknown variable declaration");
			}
		 }
	  } /* Finished processing the declaration section of the computation */
	  return env;
   } /* initializeEnvironment() */   
   
   /**
	* Make this computation into a relation on the current 'data'
	* @author Patrick Baker
	* modified by Yi Zheng
	*/
   private Relation makeRelation(String name, Environment env)
   throws InterpretError
   {  
	  Relation r = new Relation(name, domains, RELATION, numtuples,
								numattrs, numsortattrs, null, declEnv);//should be callEnv

	  r.data = env.data;
	 //remove duplicates added by Yi Zheng April 6, 2002
	  r.save();
	  r = r.project(domains,r.name); 
	  r.unload();
	  return r;
   }   /* makeRelation */
   
   /**
    * replace join attributes with entries in nameMapping
    * @author Yi Zheng June 20, 2002
    */
    private Domain[] replaceJoinAttrs(Domain[] doms, Hashtable tbl)
		throws InterpretError
    {
    	Domain[] ret = new Domain[doms.length];
    	for(int i = 0; i < doms.length; i++)
    	{
    		if(isDomainIn(doms[i], joinAttrs))
    		{
	    		Domain d = (Domain) tbl.get(doms[i].name);
	    		if(d == null) //should never happen
	    			throw new InterpretError(doms[i].name + " not found");
	    		ret[i] = d;
	    	}
	    	else
	    		ret[i] = doms[i];
    	}
    	return ret;
    } /* replaceJoinAttrs */
    
   /**
	* This routine is used when a computation is called as part of 
	* a selector. It parses the conditional expression in the selector
	* which provides inputs to the computation. These values are put
	* directly into the 'data' array of arrays.
	* @param   doms   used to accumulate the input domains.
	* @author Patrick Baker
	* modified by Yi Zheng
	*/
   private void parseSelectConditional(SimpleNode node, Vector inputDoms, 
									   Environment env, Environment callEnv)
   throws InterpretError
   { 
    if (node.type == OP_BIOPERATOR)
	  {  //binary operator must have two children
		 SimpleNode nodel = (SimpleNode) node.jjtGetChild(0);
		 SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
		 switch (node.opcode) //note that only AND, EQ will be used here
		 {  case OP_AND :
			   parseSelectConditional(nodel, inputDoms, env, callEnv);
			   parseSelectConditional(noder, inputDoms, env, callEnv);
			   break;
			case OP_EQ :
			   int col = -1;
			   int type = -1;
			   SimpleNode expr = null;
			   if (nodel.type == OP_IDENTIFIER)
			   {  ParamInfo pi = env.lookupParam(nodel.name, false);
				  if (pi != null)
				  {
				  	 Domain d = env.lookupDom(nodel.name, true);
				  	 if(d == null)
				  	  	throw new InterpretError("domain " + nodel.name + " does not exist");
					 inputDoms.addElement(d);
					 col = pi.index;
					 type = pi.type;
					 
					 expr = noder;
				  }else if(DEBUG)
				  {
				  }
			   }
			   		
			   if (noder.type == OP_IDENTIFIER)
			   {  ParamInfo pi = env.lookupParam(noder.name, false);
				  if (pi != null)
				  {  
				  if (expr != null)
					 {  // '=' sign is surrounded by two parameters, this is an 
						// ambiguous situation. We are forced to throw an error.
						// See last chapter of thesis for a discussion of this
						throw new InterpretError 
							  ("'=' sign in selection clause is surrounded by two"+
							   " parameters.\n This is ambiguous.");
					 }
					 else
					 {  Domain d = env.lookupDom(noder.name, true);
					    if(d == null) 
					    	throw new InterpretError("domain " + noder.name + " does not exist");
						inputDoms.addElement(d);
						col = pi.index;
						type = pi.type;
						expr = nodel;
					 }
				  }else if(DEBUG)
				  {
				  }

			   }
			   
			   if (expr == null)
			   {  throw new InterpretError
					   ("A parameter must appear on one side of the '=' sign\n"+
						"in a select clause.");
			   }
			   switch (type)
			   {  case BOOLEAN :
		//             ((byte[]) data[col])[0] = EvalExpr.evalBooleanExpr(expr, callEnv);
		//			 break; Yi Zheng commented out March 31, 2002 REVISIT
						throw new InterpretError("BOOLEAN type not implemented yet in selector of computation");
						
				  case SHORT :
					 ((short[]) data[col])[0] = EvalExpr.evalShortExpr(expr, callEnv);
					 break;
				  case INTEGER :
					 ((int[]) data[col])[0] = EvalExpr.evalIntExpr(expr, callEnv);
					 break;
				  case LONG :
					 ((long[]) data[col])[0] = EvalExpr.evalLongExpr(expr, callEnv);
					 break;
				  case FLOAT :
					 ((float[]) data[col])[0] = EvalExpr.evalFloatExpr(expr, callEnv);
					 break;
				  case DOUBLE :
					 ((double[]) data[col])[0] = EvalExpr.evalDoubleExpr(expr, callEnv);
					 break;
				  case STRING :
		//			 break; commented out by Yi Zheng March 31, 2002, REVISIT
					 ((String[]) data[col])[0] = EvalExpr.evalStringExpr(expr, callEnv);
					 break;
		//			 throw new InterpretError("STRING type not implemented yet in selector of computation");
					 
				  case IDLIST : //Yi Zheng REVISIT: //CONTINUE MAY 30
				  // try this case: domain a intg;
				  //				domain r1, r2(a);
				  //				comp C(r1,r2) is
				  //				{ r2 <- r1;};
				  //				relation rin(a) <- {(1),(2)};
				  //				rout <- C[rin];
					 //init hashTable to satisfy evaluateTLExpression REVISIT:
					 Hashtable ht = new Hashtable();
					 ht.put("#env3", env);
					 Relation r = interpreter.evaluateTLExpression(
												 expr,"pat_temp",ht, env);
					 if (r == null)
					 {  throw new InterpretError(
						"Can't evaluate expression in select clause.");
					 }
					 
					 long surr = Surrogate.next();
					 ((long[]) data[col])[0] = surr;
					 Relation dom = env.lookupRel("."+nodel.name, true); 
			//		 Relation dom = env.lookupRel("."+expr.name, true);
					 if (dom == null)
					 {  throw new 
						InterpretError("Can't find ."+nodel.name); 
					 }
					 dom.insert(surr, r);
					 break;
				  case COMP :
			//		 break;	commented out by Yi Zheng March 31, 2002
					throw new InterpretError("COMP not implemented as a legal type for comparison in selector of computation");
					
				  case TEXT :
				  case STMT :
				  case EXPR :
					 throw new
					 InterpretError("Types 'text', 'stmt', and 'expr' " +
									"are not yet supported in computations");
				  default :
					 throw new InterpretError("Unknown type");
			   } /* switch (type) */
			   break;
			default :
			   throw new 
			   InterpretError("Invalid selection clause for computations");
		 } /* switch (node.opcode) */
	  }
	  //added by Yi Zheng April 8, 2002 to bug fix: aComp[];
	  else if(node.type == OP_LITERAL && node.opcode == OP_BOOLEAN)
	  {
	  	byte val = ((Byte)node.getObject()).byteValue();
	  	if (val == 1)
	  		return;	//no need to update inputdoms
	  	else //IMPOSSIBLE
	  		throw new InterpretError("Invalid selection clause for computations");
	 }
	  else //node.type != OP_BIOPERATOR && node.type != OP_LITERAL
	  {  throw new 
		 InterpretError("Invalid selection clause for computations");
	  }
   } /* parseSelectConditional() */   
   
              
   /**
	* invoked by applyInOut to put input call parameters into LocalInfo objects
	* Remark: currently call parameters have to be either literal or an identifier
	* that is a domain of the enclosing environment;
	* future work might modify this method to handle expressions in call parameter as well
	* @author Patrick Baker
	* modified by Yi Zheng
	*/
   private int processCallParams(SimpleNode callParams, String prefix,
								 Environment env, Environment callEnv)
   throws InterpretError
   {
	  //Put a copy of all inputs into the new frame of the environment
	  //Calculate the block type at the same time
	  int blockType = 0;
	  int numCallParams = (callParams == null) ? 0 
										  : callParams.jjtGetNumChildren();


	  //Some semantic checks...
	  //Check that the right number of parameters were used in the call
	  if (numCallParams != numattrs)
	  {  throw new InterpretError
			   ("Wrong number of parameters provided in computation call");
	  }

	  SimpleNode body;
	  if (tree.jjtGetNumChildren() == 3)
	  { 
	    body = (SimpleNode) tree.jjtGetChild(2);
	  }
	  else //(node.jjtGetNumChildren() == 2)
	  {  //Empty parameter list in this case
		 body = (SimpleNode) tree.jjtGetChild(1);
	  }
	  SimpleNode decl = (SimpleNode) body.jjtGetChild(0);
	  if( decl.type == OP_COMPVARDECL)
	  {  if (decl.jjtGetNumChildren() != 0)
		 {  //The optional declaration section is in fact present
		//	throw new InterpretError(
		//			  "Computation has local and/or state variables.\n" +
		//			  "It may not be called with the in/out syntax");
		//commented out by Yi Zheng April 18, 2002
		 }
	  }

	  for(int i=0; i<numCallParams; i++)
	  {  
	  	 Domain nd = null;
	  	 SimpleNode cp = (SimpleNode) callParams.jjtGetChild(i);
		 if (!cp.isBitSet(0))
		 {  //cp is an input parameter
			Object val = null;
			SimpleNode literal = null;

			blockType |= 1 << i; //adjust blocktype
			
			//weizhong
			if (cp.jjtGetNumChildren() == 0)
			{
			//revisit: Is it possible to be here?
			  
			}
			
			else if (cp.jjtGetNumChildren() == 1)
			{ literal = (SimpleNode) cp.jjtGetChild(0);
			  LocalInfo l = null;

			  switch(literal.opcode)
			  { case OP_BOOLEAN :
		//		   if (literal != null) //this is not necessary since we are already here, otherwise switch will throw exception
				   val = new Boolean(EvalExpr.evalBooleanExpr(literal, env));
				   l = new LocalInfo(BOOLEAN, val);
				   env.put(prefix+domains[i].name, l);
				   break;
				case OP_SHORT :
				   val = new Short(EvalExpr.evalShortExpr(literal, env));
				   l = new LocalInfo(SHORT, val);
				   env.put(prefix+domains[i].name, l);
				   break;
				case OP_INTEGER :
				   val = new Integer(EvalExpr.evalIntExpr(literal, env));
				   l = new LocalInfo(INTEGER, val);
				   env.put(prefix+domains[i].name, l);
				   break;
				case OP_LONG :
				   val = new Long(EvalExpr.evalLongExpr(literal, env));
				   l = new LocalInfo(LONG, val);
				   env.put(prefix+domains[i].name, l);
				   break;
				case OP_FLOAT :
				   val = new Float(EvalExpr.evalFloatExpr(literal, env));
				   l = new LocalInfo(FLOAT, val);
				   env.put(prefix+domains[i].name, l);
				   break;
				case OP_DOUBLE :
				   val = new Double(EvalExpr.evalDoubleExpr(literal, env));
				   l = new LocalInfo(DOUBLE, val);
				   env.put(prefix+domains[i].name, l);
				   break;
				case OP_STRING :
				   val = EvalExpr.evalStringExpr(literal, env);
				   l = new LocalInfo(STRING, val);
				   env.put(prefix+domains[i].name, l);
				   break;
		//		case OP_TEXT : commented out and moved to end by Yi Zheng March 24, 2002
		//		case OP_STMT :
		//		case OP_EXPR :
				//TEXT,STMT and EXPR are not implemented. Some code
				//for these will go here when they are implemented.
				//For now, it is impossible to actually be here.
				case OP_IDENTIFIER: //Yi Zheng March 24, 2002, Can we assume it's a relation? May 15, NO!!! 
				// weizhong //Yi Zheng, not sure what these 3 lines do? 
				
				
			//		val = new String(literal.name);
			//		l = new LocalInfo(IDLIST, val);
			//		env.put(prefix+domains[i].name,	l);
			// MOVED to inside if (id.kind == REL)
			
			
			
						// start Zongyan's additioin   Aug 18 ,2002
		     boolean aldatpFlag = false;
			 if (cp.url != null) {
			   literal.url = cp.url;
			   aldatpClient.request(literal,env);  // need to check response and return it ^_^
	           if (aldatpClient.rel ==null)
	             	throw new InterpretError( aldatpClient.rspCode);
	  
	        //  if (aldatpClient.rel.rvc == 17) { // comp
	        //	 Computation rmtRel = new Computation(aldatpClient.rel.tree,env); //^_^ which env?
	        //	 literal.name = rmtRel.name;
			//	 callEnv.put(rmtRel);
	        //  }
			//  else 	{// regular relation
	        	 Relation rmtRel = aldatpClient.rel; 
	             literal.name = rmtRel.name;
				 callEnv.put(rmtRel);
				 aldatpFlag = true;
			//  }
			 }
				    
		    // end Zongyan's addition

					//added by Yi Zheng March 24, 2002, assuming literal is a relation's name
					//add it to env 
					IDInfo id = callEnv.lookup(literal.name, true);
					if(id == null)
			  		{  throw new InterpretError
						   ("Input '"+literal.name+"' does not exist");
			  		}
			  	
			  		if (id.kind == REL)
			  		{
			  			Relation r = ((RelInfo) id).rel;

			  			val = new String(literal.name);
						if((r.rvc == RELATION)||(r.rvc == VIEW))
							l = new LocalInfo(IDLIST, val);
						else
							l = new LocalInfo(COMP, val);//Yi Zheng June 17, 2002
						env.put(prefix+domains[i].name,	l);

			  			
	  			  		Relation newr;
				 		if(r.rvc == RELATION || r.rvc == VIEW)
				 		{
				 			if((r.rvc == RELATION) && r.isLoaded())
				 				r.save(); 
				 			newr = r.duplicate(true);
				 		}
				 		else //must be COMPUTATION
				 		{  //newr = ((Computation) r).duplicate(); //original code
				 		
				 		  //modified by Yi Zheng June 17, 2002
				 		  newr = new Computation(((SimpleNode)r.tree).jjtDeepDuplicate(), env);
				 		  //end of modification June 17, 2002
				 		}
				 	//	newr.name = prefix+domains[i].name; commented out by Yi Zheng
				 		//added by Yi Zheng March 25, 2002 as a bug fix
						newr.name = prefix+domains[i].name;
				 		if((newr.data == null) && (newr.rvc == RELATION)) newr.load();
				 		env.put(newr);
				 		if(newr.rvc == RELATION) newr.save(); //added by Yi Zheng March 25, 2002 for put will delete the data file first

						l.relVal = newr; //added by Yi Zheng May 17, 2002
			  		}
			  		else if (id.kind == LOCAL) //added by Yi Zheng May 20, 2002
			  		{
			  			val = new String(literal.name);
						if(id.type == COMP)
							l = new LocalInfo(COMP, val); //Yi Zheng June 17, 2002
						else 
							l = new LocalInfo(IDLIST, val);
						env.put(prefix+domains[i].name,	l);

			  			
	  			  		Relation newr;
				 		Relation r = ((LocalInfo) id).relVal;
				 		if(r.rvc == RELATION || r.rvc == VIEW)
				 		{
				 			if((r.rvc == RELATION) && r.isLoaded())
				 				r.save(); 
				 			newr = r.duplicate(true);
				 		}
				 		else //must be COMPUTATION
				 		{  //newr = ((Computation) r).duplicate(); //Yi Zheng June 17, 2002
				 		 //modified by Yi Zheng June 17, 2002
				 		  newr = new Computation(((SimpleNode)r.tree).jjtDeepDuplicate(), env);
				 		  //end of modification June 17, 2002

				 		}
						newr.name = prefix+domains[i].name;
				 		if((newr.data == null) && (newr.rvc == RELATION)) newr.load();
				 		env.put(newr);
				 		if(newr.rvc == RELATION) newr.save(); //added by Yi Zheng March 25, 2002 for put will delete the data file first

						l.relVal = newr; //added by Yi Zheng May 17, 2002
			  		}
			  		else if (id.kind == STATE) //added by Yi Zheng May 20, 2002
			  		{
			  			val = new String(literal.name);
			  			if(id.type == COMP)
			  				l = new LocalInfo(COMP, val); //Yi Zheng June 17, 2002
			  			else
							l = new LocalInfo(IDLIST, val);
						env.put(prefix+domains[i].name,	l);

			  			
	  			  		Relation newr;
				 		Relation r = ((StateInfo) id).relVal;
				 		if(r.rvc == RELATION || r.rvc == VIEW)
				 		{
				 			if((r.rvc == RELATION) && r.isLoaded())
				 				r.save(); 
				 			newr = r.duplicate(true);
				 		}
				 		else //must be COMPUTATION
				 		{  //newr = ((Computation) r).duplicate(); //original code
				 		 //modified by Yi Zheng June 17, 2002
				 		  newr = new Computation(((SimpleNode)r.tree).jjtDeepDuplicate(), env);
				 		  //end of modification June 17, 2002
				 		
				 		}
						newr.name = prefix+domains[i].name;
				 		if((newr.data == null) && (newr.rvc == RELATION)) newr.load();
				 		env.put(newr);
				 		if(newr.rvc == RELATION) newr.save(); //added by Yi Zheng March 25, 2002 for put will delete the data file first

						l.relVal = newr; //added by Yi Zheng May 17, 2002
			  		}
			  		//added by Yi Zheng, July 7, 2002
			  		//to implement BANK
			  		else if (id.kind == DOM)
			  		{
			  			Domain idDom = ((DomInfo)id).dom;
			  			if(idDom.tree != null)
			  			{
			  				SimpleNode declNode = (SimpleNode) (((SimpleNode)idDom.tree).jjtGetChild(0));
			  				switch(idDom.type)
			  				{
			  					case INTEGER:
			  						val = new Integer(EvalExpr.evalIntExpr(declNode, callEnv));
			  						break;
			  					case SHORT:
			  						val = new Short(EvalExpr.evalShortExpr(declNode, callEnv));
			  						break;
			  					case BOOLEAN:
			  						val = new Boolean(EvalExpr.evalBooleanExpr(declNode, callEnv));
			  						break;
			  					case LONG:
			  						val = new Long(EvalExpr.evalLongExpr(declNode, callEnv));
			  						break;
			  					case FLOAT:
			  						val = new Float(EvalExpr.evalFloatExpr(declNode, callEnv));
			  						break;
			  					case DOUBLE:
			  						val = new Double(EvalExpr.evalDoubleExpr(declNode, callEnv));
			  						break;
			  					case STRING:
			  						val = EvalExpr.evalStringExpr(declNode, callEnv);
			  						break;
			  					default:
			  						throw new InterpretError("to be implemented");
			  				}
			  				l = new LocalInfo(idDom.type, val);
							env.put(prefix+domains[i].name,	l);
						}
						else throw new InterpretError("processCallParams: to be implemented");
			  			
			  		}
			  		//end of Yi's addition July 7, 2002
			  		else
			  		{  //REVISIT: not implemented.. prarminfo ...
			  			throw new InterpretError("Computation: processCallParams: OP_IDENTIFIER: not implemented yet");
			  		}
			  		//end of Yi Zheng's addition
			  		
			  		//added by Zongyan
                    if (aldatpFlag == true)
					    callEnv.removeRel(literal.name, false);

					break;
				case OP_TEXT : 
				case OP_STMT :
				case OP_EXPR :
				//TEXT,STMT and EXPR are not implemented. Some code
				//for these will go here when they are implemented.
				//For now, it is impossible to actually be here.
					
				default :
				   throw new InterpretError
						 ("Invalid or not implemented type call parameter");
			  }
			}
			else //BUG REVISIT: how can we ever come here? Doesn't one param
			//have only one associated child? -- Yi Zheng March 23, 2002
			{ 
			IDInfo id = callEnv.lookup(cp.name, true);
			  if(id == null)
			  {  throw new InterpretError
						   ("Input '"+cp.name+"' does not exist...");
			  }
			  if (id.kind == REL)
			  {  Relation newr;
				 Relation r = ((RelInfo) id).rel;
				 if(r.rvc == RELATION || r.rvc == VIEW)
				 {  newr = r.duplicate(true);
				 }
				 else //must be COMPUTATION
				 {  newr = ((Computation) r).duplicate();
				 }
				 newr.name = prefix+domains[i].name;
				 env.put(newr);
			  }
			  else
			  {  //REVISIT: not implemented.. prarminfo localinfo...
			  	//added by Yi Zheng April 18, 2002
			  	throw new InterpretError("Not implemented");
			  }
			}
		 } /* case for an input parameter */
	  } /* for loop over call parameters*/
	  return blockType;
   } /* processCallParams */ 
     
   /**
	* This procedure is called only by applyInOut()
	* It renames the identifiers in the syntax tree that are parameters
	* of the computation by adding a 'prefix'. This is needed because
	* there is no in-RAM evaluator. We can tell if  an idenitifer is 
	* a parameter by checking that the inoutStatus is -2. This has to be
	* set up before calling this routine.
	* Special care must be taken to respect the scoping rules.
	* @author Patrick Baker
	* modified by Yi Zheng
	*/
   private void renameIDs(SimpleNode node, String prefix, Environment env)
   throws InterpretError
   {  switch(node.type)
	  {  case OP_IDENTIFIER:
			IDInfo id = env.lookup(node.name, true);
		//	if (id != null)  Yi Zheng May 14, 2002
			if (id != null && id.kind == PARAMETER)
			{  node.name = prefix+node.name;
			}
			break; 
		 case OP_DECLARATION ://REVISIT: don't we have to make sure this is compuation declaration?
			 if(node.opcode == OP_COMPUTATION) //if statement (not body) added by Yi Zheng March 24, 2002
			 {
				//commented out July 24, 2002
				//Environment newEnv = new Environment(env);
				//initializeEnvironment(node, newEnv, null);
				int numChildren = node.jjtGetNumChildren();
				if (numChildren == 3)
				{ renameIDs((SimpleNode) node.jjtGetChild(2), prefix, env);
				}
				else
				{ renameIDs((SimpleNode) node.jjtGetChild(1), prefix, env);
				}
			}
			else if(node.opcode == OP_VIEW) //addition by Yi Zheng May 20, 2002
			{
				SimpleNode node_init = null;
				SimpleNode node_is = null;
				SimpleNode node_target = null;
				
				if(node.isBitSet(0))
				{
					node_init = (SimpleNode) node.jjtGetChild(1);
					node_is = (SimpleNode) node.jjtGetChild(2);
				}
				else
					node_is = (SimpleNode) node.jjtGetChild(1);
				if(node.isBitSet(1))
				{
					if(node.isBitSet(0))
						node_target = (SimpleNode) node.jjtGetChild(3);
					else
						node_target = (SimpleNode) node.jjtGetChild(2);
				}
				if(node_init != null)
					renameIDs(node_init, prefix, env);
				renameIDs(node_is, prefix, env);
				if(node_target != null)
					renameIDs(node_target, prefix, env);
			}
			else if(node.opcode == OP_LET) //addition by Yi Zheng May 20, 2002
			{
				SimpleNode node_init = null;
				SimpleNode node_be = null;
				if(node.isBitSet(0))
				{
					node_init = (SimpleNode) node.jjtGetChild(1);
					node_be = (SimpleNode) node.jjtGetChild(2);
				}
				else
					node_be = (SimpleNode) node.jjtGetChild(1);
				if(node_init != null)
					renameIDs(node_init, prefix, env);
				renameIDs(node_be, prefix, env);	
			}
			break;
		 default :
		   for(int i=0; i<node.jjtGetNumChildren(); i++)
		   {  SimpleNode ch = (SimpleNode) node.jjtGetChild(i);
			  renameIDs(ch, prefix, env);
		   }
	  }
   } /* renameIDs */   
   
   /** invoked by applyIjoin() and applySelect()
    * @author Patrick Baker
    * modified by Yi Zheng
    */
   private void run(CompBlock block, Environment env)
   throws InterpretError
   {  int col, type, opcode;
	//added by Yi Zheng April 14, 2002
	int oldNumTuples = numtuples;

	//  for(env.row=0; env.row<numtuples; env.row++)
	for(env.row=0; env.row<oldNumTuples; env.row++)
	 
	 { 
	     for(int stmtNo=0; stmtNo<block.numStatements; stmtNo++)
		 { 
		  SimpleNode stmt = block.statements[stmtNo];
			runSingleStmt(stmt, env,null); //last argument added by Zongyan Nov 25, 2004
		 }
	  }
   }   /* run */

	/**
	 * invoked by applyIjoin for named join syntax
     * @author Yi Zheng June 20, 2002
     */
   private void renameParamsInTree(SimpleNode tree, Hashtable tbl)
   {
   	for(Enumeration e = tbl.keys(); e.hasMoreElements();)
   	{
   		String paramName = (String) e.nextElement();
   		String newName = ((Domain) tbl.get(paramName)).name;
   		renameParamInTree(tree, paramName, newName);
   	}
   } /* renameParamsInTree */

   /**
    * rename a single parameter
    * @author Yi Zheng   
    */
   private void renameParamInTree(SimpleNode tree, String oldName, String newName)
   {
   	if(tree.type == OP_IDENTIFIER || tree.type == OP_PARAMETER)
   	{
   		if(tree.name.compareTo(oldName) == 0)
   			tree.name = newName;
   	}
   	else
   	{
   		for(int i = 0; i < tree.jjtGetNumChildren(); i++)
   			renameParamInTree((SimpleNode)tree.jjtGetChild(i), oldName, newName);
    }
   } /* renameParamInTree */
   		
   /**
    * run constraint mode: when all parameters are input
    * note that a block of a required type is only found for exact match
    * if all parameters are input, the first block is returned
    * all other cases are error cases where block is not found
    * @author Yi zheng
    */
   private void runConstraint(CompBlock block, Environment env, Vector failRows)
   throws InterpretError
   {  int col, type, opcode;
	  for(env.row=0; env.row<numtuples; env.row++)
	  { 
	     for(int stmtNo=0; stmtNo<block.numStatements; stmtNo++)
		 {  SimpleNode stmt = block.statements[stmtNo];
			runSingleStmtConstraint(stmt, env, failRows);
		 }
	  }
   }  /* runConstraint */
   
   /**
    * update env.data to delete rows that failed constraint testing and make a relation
    * of the final result
    * @author Yi Zheng
    */
   private Relation makeRelationConstraint(String name, Environment env, Vector failRows)
   throws InterpretError
   {
   	int size = numtuples - failRows.size();	//is duplicates possible? REVISIT
   	if(size < 0)
	   	throw new InterpretError("makeRelationConstraint: negative size");

	//create the data here
	createDataArray(size);
		
	int i = 0, j = 0;
	for (int k = 0; k < numattrs; k++)
	{
		
			while (i < numtuples)
			{
				Integer In = new Integer(i);
				if(!failRows.contains(In))
				{
					int stmtType = domains[k].type;
							  switch(stmtType) 
		  //REVISIT
		  //Yi Zheng March 24, 2002, we are assuming that the right side has the same type as the left side
		  {  case BOOLEAN :
				//REVISIT: implement this case..
				//added by Yi Zheng March 24, 2002
				throw new InterpretError("BOOLEAN type for statement not implemented yet");
			//	break;
			 case SHORT : ((short[])data[k])[j] = ((short[])env.data[k])[i];
				break;
			 case INTEGER :
				((int[])data[k])[j] = ((int[])env.data[k])[i];
				break;
			 case IDLIST:
			 case COMP:
			 case LONG :
				((long[])data[k])[j] = ((long[])env.data[k])[i];
				break;
			 case FLOAT :
				((float[])data[k])[j] = ((float[])env.data[k])[i];
				break;
			 case DOUBLE :
				((double[])data[k])[j] = ((double[])env.data[k])[i];
				break;
			 case STRING :
		//		break; commented out by Yi Zheng March 31, 2002
				throw new InterpretError("STRING not implemented for assignment");
			 case TEXT :
			 case STMT :
			 case EXPR :
				throw new
				InterpretError("Types 'comp', 'text', 'stmt', and 'expr' " +
							   "are not yet supported in computations");
			 default :
				throw new InterpretError("Unknown type");
		  }
					j++;
				}
				i++;
			}
			
   	 
   	}
   	  Relation r = new Relation(name, domains, RELATION, size,
								numattrs, numsortattrs, null, declEnv);
	  r.data = data;
	  r.save();
	  r.unload();
	  this.unload();
	  return r;
   } /* makeRelationConstraint */
   
   /**
   	* in constraint mode, mark the input rows to be deleted 
   	* only the simplest cases are implemented
   	* @author Yi Zheng
   	*/
   	private void runSingleStmtConstraint(SimpleNode stmt, Environment env, Vector failRows)
   throws InterpretError
   {  int col;
	  int type = stmt.type;
	  int opcode = stmt.opcode;
	  int numChildren = stmt.jjtGetNumChildren();
	boolean fit;
	switch(type)
    {  /* switch */
	  case OP_ASSIGNMENT :
	   if(numChildren == 2)
	   {  SimpleNode left = (SimpleNode) stmt.jjtGetChild(0);
		  SimpleNode expr = (SimpleNode) stmt.jjtGetChild(1);
		  IDInfo id = env.lookup(left.name, true);
		  if (id == null)
		  {  //this must be an on the fly relation declaration
			 //REVISIT: Is this right
				 Relation r=new Relation(left.name, new Domain[0], 0, 0, 0, env);

			 id = new RelInfo(r);
			 env.put(left.name, id);//will later be replaced
		  }

		  int stmtType = RELIX_VOID;
		  if (id.kind == NR)
		  {  stmtType = ((NRInfo) id).type;
		  }
		  else if (id.kind == PARAMETER)
		  {  
		    stmtType = ((ParamInfo) id).type;
		  }
		  else if (id.kind == LOCAL)
		  {  //REVISIT: Should not allow type IDList for locals
			 stmtType = ((LocalInfo) id).type;
		  }
		  else if (id.kind == STATE)
		  { 
		    stmtType = ((StateInfo) id).type;
		  }
		  else if (id.kind == DOM)
		  {
		    throw new 
			 InterpretError("Illegal: Cannot assign to a domain");
		  }
		  else if (id.kind == REL)
		  {  stmtType = IDLIST;
		  }
		  else
		  {  //should never happen
			 throw new InterpretError("Internal:variable has no type");
		  }
		  switch(stmtType) 
		  //REVISIT
		  //Yi Zheng March 24, 2002, we are assuming that the right side has the same type as the left side
		  {  case BOOLEAN :
				//REVISIT: implement this case..
				//added by Yi Zheng March 24, 2002
				throw new InterpretError("BOOLEAN type for statement not implemented yet");
			//	break;
			 case SHORT :
				if (stmt.opcode == OP_ASSIGNADD)
				{  throw new 
				   InterpretError("Illegal assign add operation");
				}
				short shortResult = EvalExpr.evalShortExpr(expr, env);
				if(id.kind == NR)
				{  throw new InterpretError("to be implemented");
				   
				}
				else if(id.kind == PARAMETER)
				{  Environment e = ((ParamInfo) id).env;
				   col = ((ParamInfo) id).index;
				   fit = (((short[]) e.data[col])[e.row] == shortResult);
				   if (!fit) failRows.addElement(new Integer(e.row));
				}
				else if(id.kind == LOCAL)
				{  ((LocalInfo) id).intVal = shortResult;
				}
				else if(id.kind == STATE)
				{
					//added by Yi Zheng April 23, 2002
					if(!isIjoin)
					 ((StateInfo) id).intVal = shortResult;
					else
					{
						col = ((StateInfo)id).index;
						((short[]) env.data[col])[env.row] = shortResult;
					}	
				}
				else
				{  throw new 
				   InterpretError("Illegal kind for identifier "+left.name);
				}
				
				break;
			 case INTEGER :
				if (stmt.opcode == OP_ASSIGNADD)
				{  throw new 
				   InterpretError("Illegal assign add operation");
				}
				
				int intResult = EvalExpr.evalIntExpr(expr, env);

				if(id.kind == NR)
				{  throw new InterpretError("not implemented");
				}
				else if(id.kind == PARAMETER)
				{  
				   Environment e = ((ParamInfo) id).env;
				   col = ((ParamInfo) id).index;
				   fit = (((int[]) e.data[col])[e.row] == intResult);
				   if (!fit) failRows.addElement(new Integer(e.row));
				}
				else if(id.kind == LOCAL)
				{  ((LocalInfo) id).intVal = intResult;
				}
				else if(id.kind == STATE)
				{ 
					if(!isIjoin)
						((StateInfo) id).intVal = intResult;
					else
					{
						col = ((StateInfo)id).index;
						((int[]) env.data[col])[env.row] = intResult;
					}
				}
				else
				{  throw new 
				   InterpretError("Illegal kind for identifier "+left.name);
				}
				

				break;
			 case LONG :
				if (stmt.opcode == OP_ASSIGNADD)
				{  throw new 
				   InterpretError("Illegal assign add operation");
				}
				long longResult = EvalExpr.evalLongExpr(expr, env);
				if(id.kind == NR)
				{  throw new InterpretError("not implemented");
				}
				else if(id.kind == PARAMETER)
				{  Environment e = ((ParamInfo) id).env;
				   col = ((ParamInfo) id).index;
				   fit = (((long[]) e.data[col])[e.row]== longResult);
				   if (!fit) failRows.addElement(new Integer(e.row));
				}
				else if(id.kind == LOCAL)
				{  ((LocalInfo) id).longVal = longResult;
				}
				else if(id.kind == STATE)
				{ 
					if(!isIjoin)
					((StateInfo) id).longVal = longResult;
					else
					{
						col = ((StateInfo)id).index;
						((long[]) env.data[col])[env.row] = longResult;
					}

				}
				else
				{  throw new 
				   InterpretError("Illegal kind for identifier "+left.name);
				}
				break;
			 case FLOAT :
				if (stmt.opcode == OP_ASSIGNADD)
				{  throw new 
				   InterpretError("Illegal assign add operation");
				}
				float floatResult = EvalExpr.evalFloatExpr(expr, env);
				if(id.kind == NR)
				{  throw new InterpretError("not implemented");
				}
				else if(id.kind == PARAMETER)
				{  Environment e = ((ParamInfo) id).env;
				   col = ((ParamInfo) id).index;
				   fit = (((float[]) e.data[col])[e.row] == floatResult);
				   if (!fit) failRows.addElement(new Integer(e.row));
				}
				else if(id.kind == LOCAL)
				{  ((LocalInfo) id).floatVal = floatResult;
				}
				else if(id.kind == STATE)
				{ if(!isIjoin)
					((StateInfo) id).floatVal = floatResult;
					else
					{
						col = ((StateInfo)id).index;
						((float[])env.data[col])[env.row] = floatResult;
					}
				}
				else
				{  throw new 
				   InterpretError("Illegal kind for identifier "+left.name);
				}
				break;
			 case DOUBLE :
				if (stmt.opcode == OP_ASSIGNADD)
				{  throw new 
				   InterpretError("Illegal assign add operation");
				}
				double doubleResult = EvalExpr.evalDoubleExpr(expr, env);
				if(id.kind == NR)
				{  throw new InterpretError("not implemented");
				}
				else if(id.kind == PARAMETER)
				{  Environment e = ((ParamInfo) id).env;
				   col = ((ParamInfo) id).index;
				   fit = (((double[]) e.data[col])[e.row] == doubleResult);
				   if (!fit) failRows.addElement(new Integer(e.row));
				}
				else if(id.kind == LOCAL)
				{  ((LocalInfo) id).doubleVal = doubleResult;
				}
				else if(id.kind == STATE)
				{ if(!isIjoin)
					((StateInfo) id).doubleVal = doubleResult;
					else
					{
						col = ((StateInfo)id).index;
						((double[])env.data[col])[env.row] = doubleResult;
					}
				}
				else
				{  throw new 
				   InterpretError("Illegal kind for identifier "+left.name);
				}
				break;
			 case STRING :
		//		break; commented out by Yi Zheng March 31, 2002
		//	Following case added by Yi Zheng April 13, 2002
				if (stmt.opcode == OP_ASSIGNADD)
				{  throw new 
				   InterpretError("Illegal assign add operation");
				}
				String strResult = EvalExpr.evalStringExpr(expr, env);
				if(id.kind == NR)
				{  throw new InterpretError("not implemented");
				}
				else if(id.kind == PARAMETER)
				{  Environment e = ((ParamInfo) id).env;
				   col = ((ParamInfo) id).index;
				   fit = ((((String)(((String[]) e.data[col])[e.row])).compareTo(strResult)) == 0);
				   if (!fit) failRows.addElement(new Integer(e.row));
				}
				else if(id.kind == LOCAL)
				{  ((LocalInfo) id).stringVal = strResult;
				}
				else if(id.kind == STATE)
				{ if(!isIjoin)
					 ((StateInfo) id).stringVal = strResult; //Yi Zheng April 18, 2002
					else
					{
						col = ((StateInfo) id).index;
				   		((String[]) env.data[col])[env.row] = strResult;
					}
				}
				else
				{  throw new 
				   InterpretError("Illegal kind for identifier "+left.name);
				}
				break;
			 case IDLIST :
			 
			 case COMP :
		//		break; commented out by Yi Zheng March 31, 2002
			 case TEXT :
			 case STMT :
			 case EXPR :
				throw new
				InterpretError("Types 'IDLIST' 'comp', 'text', 'stmt', and 'expr' " +
							   "are not yet supported in computations with constraints");
			 default :
				throw new InterpretError("Unknown type");
		  }
	   }
	   else if (numChildren == 4)
	   {  //REVISIT: assignment with 4 children
	   //added by Yi Zheng March 24, 2002
	   throw new InterpretError("Assignment with renaming not implemented yet");
	   }
	   else
	   {  throw new  //should never happen
		  InterpretError("Invalid assignment statement");
	   }
	   break;
	case OP_COMMAND :
	   throw new 
	   InterpretError("Commands not supported in constraint mode in computations");
	case OP_DECLARATION :
			 throw new 
			 InterpretError("Not implemented yet in constraint mode in computation");
	case OP_STATEMENT :
	   switch(opcode) 
	   {  case OP_CONDITIONAL :
			 SimpleNode ifNode, thenNode, elseNode;
			 ifNode = (SimpleNode) stmt.jjtGetChild(0);
			 thenNode = (SimpleNode) stmt.jjtGetChild(1);
	 
			 if (EvalExpr.evalBooleanExpr(ifNode, env))
			 {  runSingleStmtConstraint(thenNode, env, failRows);
			 }
			 else if(numChildren == 3)
			 {  elseNode = (SimpleNode) stmt.jjtGetChild(2);
				runSingleStmtConstraint(elseNode, env, failRows);
			 }
			 break;
		  case OP_STMTBLOCK :
			 for(int stmtNo=0; stmtNo<stmt.jjtGetNumChildren(); stmtNo++)
			 {  runSingleStmtConstraint((SimpleNode) stmt.jjtGetChild(stmtNo), env, failRows);
			 }
			 break;
		  case OP_SEQUENCE :
		  case OP_PARALLEL :
		  case OP_CHOICE :
		  case OP_COMPCALL :
		  case OP_FORLOOP :
		  case OP_WHILELOOP :
		  case OP_EXIT :
		  case OP_DEADLOCK :
		  case OP_EXEC :
			 throw new 
			 InterpretError("Not implemented yet");
	   }
	   break;
	case OP_UPDATE :
	   throw new 
	   InterpretError("Not implemented yet");
	default :
	   throw new 
	   InterpretError("Illegal statement");
 }  /* switch */
   } /* runSingleStmtConstraint() */   

  /**
   * method called by run
   * @author Patrick Baker
   * modified by Yi Zheng
   */

   private void runSingleStmt(SimpleNode stmt, Environment env, Environment callEnv) //not used by Weizhong's applyInOut
   throws InterpretError   // last argument added by Zongyan Nov 25, 2004
   { 
   //addition by Yi Zheng May 20, 2002
    //remember the old interpreter.env
    Environment oldEnv = interpreter.env;
    //set interpreter.env to env
    //this is necessary due to the fact that some methods in Interpreter
    //uses the object variable env, instead of the one passed in as a parameter
    interpreter.env = env;
    //end of addition May 20, 2002

	try{    
    	int col;
	  	int type = stmt.type;
	  	int opcode = stmt.opcode;
	  	int numChildren = stmt.jjtGetNumChildren();
                String prefix = "__#"+name;
                int j;


	  switch(type)
 	  {  /* switch */
		case OP_ASSIGNMENT :
		   if(numChildren == 2)
		   {  SimpleNode left = (SimpleNode) stmt.jjtGetChild(0);
			  SimpleNode expr = (SimpleNode) stmt.jjtGetChild(1);
			  IDInfo id = env.lookup(left.name, true);
			  if (id == null)
			  {  //this must be an on the fly relation declaration
				 //REVISIT: Is this right
				 Relation r=new Relation(left.name, new Domain[0], 0, 0, 0, env);
				 id = new RelInfo(r);
				 env.put(left.name, id);//will later be replaced
			  }
	
			  int stmtType = RELIX_VOID;
			  if (id.kind == NR)
			  {  
			  	stmtType = ((NRInfo) id).type;
			  }
			  else if (id.kind == PARAMETER)
			  { 
			    stmtType = ((ParamInfo) id).type;
			  }
			  else if (id.kind == LOCAL)
			  {  
				 stmtType = ((LocalInfo) id).type;
			  }
			  else if (id.kind == STATE)
			  { 
			    stmtType = ((StateInfo) id).type;
			  }
			  else if (id.kind == DOM)
			  {  
			    throw new 
				 InterpretError("Illegal: Cannot assign to a domain");
			  }
			  else if (id.kind == REL)
			  { 
			   stmtType = IDLIST;
			  }
			  else
			  {  //should never happen
				// throw new InterpretError("Internal:variable has no type"); Yi Zheng May 15, 2002
				throw new InterpretError("LHS has illegal type: " + id.kind);
			  }
			  switch(stmtType) 
			  //REVISIT
			  //Yi Zheng March 24, 2002, we are assuming that the right side has the same type as the left side
			  {  case BOOLEAN :
					//REVISIT: implement this case..
					//added by Yi Zheng March 24, 2002
					throw new InterpretError("BOOLEAN type for statement not implemented yet");
				//	break;
				 case SHORT :
					if (stmt.opcode == OP_ASSIGNADD)
					{  throw new 
					   InterpretError("Illegal assign add operation");
					}
					short shortResult = EvalExpr.evalShortExpr(expr, env);
					if(id.kind == NR)
					{  NREnvironment nre = ((NRInfo) id).NRenv;
					   col = ((NRInfo) id).index;
					   ((short[]) nre.data[col])[nre.row] = shortResult;
					}
					else if(id.kind == PARAMETER)
					{  Environment e = ((ParamInfo) id).env;
					   col = ((ParamInfo) id).index;
					   ((short[]) e.data[col])[e.row] = shortResult;
					}
					else if(id.kind == LOCAL)
					{  ((LocalInfo) id).intVal = shortResult;
					}
					else if(id.kind == STATE)
					{ 
					  //added by Yi Zheng April 18, 2002 as a bug fix
					  //further modified on April 23, 2002
					  if(!isIjoin)
					  	((StateInfo) id).intVal = shortResult;
					  else
					  {
					  		col = ((StateInfo) id).index;
					  		((short[]) env.data[col])[env.row] = shortResult;
					  }
					}
					else
					{  throw new 
					   InterpretError("Illegal kind for identifier "+left.name);
					}
					break;
				 case INTEGER :
					if (stmt.opcode == OP_ASSIGNADD)
					{  throw new 
					   InterpretError("Illegal assign add operation");
					}
					
					int intResult = EvalExpr.evalIntExpr(expr, env);
	
					if(id.kind == NR)
					{  NREnvironment nre = ((NRInfo) id).NRenv;
					   col = ((NRInfo) id).index;
					   ((int[]) nre.data[col])[nre.row] = intResult;
					}
					else if(id.kind == PARAMETER)
					{  
					   Environment e = ((ParamInfo) id).env;
					   col = ((ParamInfo) id).index;
					   ((int[]) e.data[col])[e.row] = intResult;
					}
					else if(id.kind == LOCAL)
					{  ((LocalInfo) id).intVal = intResult;
					}
					else if(id.kind == STATE)
					{ 
						if(!isIjoin)
							((StateInfo) id).intVal = intResult; //Yi Zheng April 18, 2002
						else
						{ 	col = ((StateInfo) id).index;
							((int[]) env.data[col])[env.row] = intResult;
						}
					}
					else
					{  throw new 
					   InterpretError("Illegal kind for identifier "+left.name);
					}
					break;
				 case LONG :
					if (stmt.opcode == OP_ASSIGNADD)
					{  throw new 
					   InterpretError("Illegal assign add operation");
					}
					long longResult = EvalExpr.evalLongExpr(expr, env);
					if(id.kind == NR)
					{  NREnvironment nre = ((NRInfo) id).NRenv;
					   col = ((NRInfo) id).index;
					   ((long[]) nre.data[col])[nre.row] = longResult;
					}
					else if(id.kind == PARAMETER)
					{  Environment e = ((ParamInfo) id).env;
					   col = ((ParamInfo) id).index;
					   ((long[]) e.data[col])[e.row] = longResult;
					}
					else if(id.kind == LOCAL)
					{  ((LocalInfo) id).longVal = longResult;
					}
					else if(id.kind == STATE)
					{  if(!isIjoin)
						((StateInfo) id).longVal = longResult; //Yi Zheng April 18, 2002
						else
						{
							col = ((StateInfo)id).index;
							((long[]) env.data[col])[env.row] = longResult;
						}
	
					}
					else
					{  throw new 
					   InterpretError("Illegal kind for identifier "+left.name);
					}
					break;
				 case FLOAT :
					if (stmt.opcode == OP_ASSIGNADD)
					{  throw new 
					   InterpretError("Illegal assign add operation");
					}
					float floatResult = EvalExpr.evalFloatExpr(expr, env);
					if(id.kind == NR)
					{  NREnvironment nre = ((NRInfo) id).NRenv;
					   col = ((NRInfo) id).index;
					   ((float[]) nre.data[col])[nre.row] = floatResult;
					}
					else if(id.kind == PARAMETER)
					{  Environment e = ((ParamInfo) id).env;
					   col = ((ParamInfo) id).index;
					   ((float[]) e.data[col])[e.row] = floatResult;
					}
					else if(id.kind == LOCAL)
					{  ((LocalInfo) id).floatVal = floatResult;
					}
					else if(id.kind == STATE)
					{ 
						if(!isIjoin)
					 		((StateInfo) id).floatVal = floatResult; //Yi Zheng April 18, 2002
						else
						{
							col = ((StateInfo) id).index;
					   		((float[]) env.data[col])[env.row] = floatResult;	
					   	}
					}
					else
					{  throw new 
					   InterpretError("Illegal kind for identifier "+left.name);
					}
					break;
				 case DOUBLE :
					if (stmt.opcode == OP_ASSIGNADD)
					{  throw new 
					   InterpretError("Illegal assign add operation");
					}
					double doubleResult = EvalExpr.evalDoubleExpr(expr, env);
					if(id.kind == NR)
					{  NREnvironment nre = ((NRInfo) id).NRenv;
					   col = ((NRInfo) id).index;
					   ((double[]) nre.data[col])[nre.row] = doubleResult;
					}
					else if(id.kind == PARAMETER)
					{  Environment e = ((ParamInfo) id).env;
					   col = ((ParamInfo) id).index;
					   ((double[]) e.data[col])[e.row] = doubleResult;
					}
					else if(id.kind == LOCAL)
					{  ((LocalInfo) id).doubleVal = doubleResult;
					}
					else if(id.kind == STATE)
					{	if(!isIjoin)
							((StateInfo) id).doubleVal = doubleResult; //Yi Zheng April 18, 2002
						else
						{
						   col = ((StateInfo) id).index;
					   		((double[]) env.data[col])[env.row] = doubleResult;
					   	}
		
					}
					else
					{  throw new 
					   InterpretError("Illegal kind for identifier "+left.name);
					}
					break;
				 case STRING ://	Following case added by Yi Zheng April 13, 2002
					if (stmt.opcode == OP_ASSIGNADD)
					{  throw new 
					   InterpretError("Illegal assign add operation");
					}
					String strResult = EvalExpr.evalStringExpr(expr, env);
					if(id.kind == NR)
					{  NREnvironment nre = ((NRInfo) id).NRenv;
					   col = ((NRInfo) id).index;
					   ((String[]) nre.data[col])[nre.row] = strResult;
					}
					else if(id.kind == PARAMETER)
					{  Environment e = ((ParamInfo) id).env;
					   col = ((ParamInfo) id).index;
					   ((String[]) e.data[col])[e.row] = strResult;
					}
					else if(id.kind == LOCAL)
					{  ((LocalInfo) id).stringVal = strResult;
					}
					else if(id.kind == STATE)
					{ 
						if(!isIjoin)
					 		((StateInfo) id).stringVal = strResult; //Yi Zheng April 18, 2002
						else
						{
						  col = ((StateInfo) id).index;
					   		((String[]) env.data[col])[env.row] = strResult;
						}
					}
					else
					{  throw new 
					   InterpretError("Illegal kind for identifier "+left.name);
					}
					break;
	
				 case IDLIST :
					Relation leftRel = null;
					if (stmt.opcode == OP_ASSIGNADD)
					{ 
					  //addition by Yi Zheng May 20, 2002
					  leftRel = env.lookupRel(left.name, true);
					  if (leftRel == null)
					  	throw new InterpretError("Relation " + left.name + " not found.");
					  //end of addition May 20, 2002
					}
					//init hashTable
					//REVISIT: evalutateTLExpression should check if ht is null
					//Yi Zheng May 17, 2002: need to enhance env with state/local/params
					//before invoking evaluateTLExpression
					
					//addition by Yi Zheng May 17, 2002, to transform the tree
					//such that all occurrences of IDENTIFIER are checked for possible
					//replacement by constants found in various IDInfo objects
					
					transformExpr(expr, env);
					//end of addition May 17, 2002
	
					Hashtable ht = new Hashtable();
					ht.put("#env3", env);
					String patTemp = interpreter.nextTempName(); //Yi Zheng June 10, 2002
					Relation relResult = 
					  // interpreter.evaluateTLExpression(expr,"pat_temp",ht, env);
					  interpreter.evaluateTLExpression(expr,patTemp,ht, env); //June 10
					if(relResult.name.compareTo("") == 0) //June 10
						relResult.name = patTemp; //June 10
					if(relResult.data != null)
						relResult.save(); //June 10
					
					if(id.kind == NR)
					{ //REVISIT?? not applicable //it's possible in runSingleStmtInUpdate
						//added by Yi Zheng April 22, 2002
						throw new InterpretError("Not implemented");
					}
					else if(id.kind == PARAMETER) //never executed by applyInOut, but used by applySelect
					{ 
					   Environment e = ((ParamInfo) id).env;
					   col = ((ParamInfo) id).index;
					   long surr = Surrogate.next();
					   ((long[]) e.data[col])[e.row] = surr;
					   //We also have to update the .paramName file
					   Relation dom = env.lookupRel("."+left.name, true);
					   if (dom == null)
					   {  throw new 
						  InterpretError("Can't find ."+left.name);
					   }
					   
					   //addition by Yi Zheng May 20, 2002
					   if(stmt.opcode == OP_ASSIGNADD)
					   	relResult.assignAdd(null, leftRel, null);
						// bug fix by Zongyan Nov 29, 2004
					   	//leftRel.assignAdd(null, relResult, null);
					   //end of addition May 20, 2002
					   dom.insert(surr, relResult);
					}
					else if(id.kind == LOCAL) //for renamed input param
					{ 
					  // added by Yi Zheng May 17, 2002
					  // since this code is not invoked by Weizhong's update code
					  // we don't need to worry about NR
					  
					  //there must be a relation in the env that was put in by processCallParams
					  //so find it and assign it relResult
					  Relation rel = env.lookupRel(left.name, true);
					  if (rel == null)
					  	throw new InterpretError(left.name + " as a relation not found");
	
					 //addition by Yi Zheng May 20, 2002
					 if(stmt.opcode == OP_ASSIGNADD)
					  // bug fix by Zongyan Nov 29, 2004
					  //rel.assignAdd(null, relResult, null);
					    relResult.assignAdd(null, rel, null);
					 //end of addition May 20, 2002
	
					  //rel = relResult.duplicate(true); 
					  //rel.name = left.name;
					  //if (rel.data == null) rel.load();
					  //rel.save();
					  //((LocalInfo)id).relVal = rel;

					  Relation r; 
					  if(stmt.opcode == OP_ASSIGNADD)
					      r = rel.duplicate(true); 
                                          else 
					      r = relResult.duplicate(true); 
					  r.name = left.name;
					  if (r.data == null) r.load();
					  r.save();
					  ((LocalInfo)id).relVal = r;
                                          // end of bug fix Nov 29, 2004

					  Interpreter.deleteFile("pat_temp"); 
					  
					  // end of addition May 17, 2002
					}
					else if(id.kind == STATE)
					{ 
					  if(!isIjoin)
					  {
						  //addition by Yi Zheng May 17, 2002
						  Relation rel = env.lookupRel(left.name, true);
						  if (rel == null)
						  	throw new InterpretError(left.name + " as a relation not found");
						 //addition by Yi Zheng May 20, 2002
						 if(stmt.opcode == OP_ASSIGNADD)
						  relResult.assignAdd(null, rel, null);
                                                  // bug fix by Zongyan Nov 29, 2004
						  //rel.assignAdd(null, relResult, null);
						 //end of addition May 20, 2002
		
						  //rel = relResult.duplicate(true); 
						  //rel.name = left.name;
						  //if (rel.data == null) rel.load();
						  //rel.save();
						  //((StateInfo)id).relVal = rel;

						  Relation r;
						  if(stmt.opcode == OP_ASSIGNADD)
						      r = rel.duplicate(true); 
                                                  else
						      r = relResult.duplicate(true); 
						  r.name = left.name;
						  if (r.data == null) r.load();
						  r.save();
						  ((StateInfo)id).relVal = r;
                                                  // end of bug fix Nov 29, 2004

					//	  Interpreter.deleteFile("pat_temp"); 
						  
						  //end of addition May 17
						}else {//called from applyIjoin added by Yi Zheng June 10, 2002
						  Environment e = ((StateInfo) id).env;
						   col = ((StateInfo) id).index;
						   long surr = Surrogate.next();
						   ((long[]) e.data[col])[e.row] = surr;
						   //We also have to update the .paramName file
						   Relation dom = env.lookupRel("."+left.name, true);
						   if (dom == null)
						   {  throw new 
							  InterpretError("Can't find ."+left.name);
						   }
						   
						   //addition by Yi Zheng May 20, 2002
						   if(stmt.opcode == OP_ASSIGNADD)
                                                        //bug fix by Zongyan Nov 29, 2004
						   	relResult.assignAdd(null,leftRel , null);
						   	//leftRel.assignAdd(null, relResult, null);
						   //end of addition May 20, 2002
						   
						   //added by Yi Zheng June 17, 2002
						   //dom.insert will fail if relResult contains hidden state dom
						   //because dom is declared on public domains only
						   // dom.numattrs != relResult.numattrs + 1
						   //so here we need to add state domains to dom if dom is still empty
						   //i.e. the state is being assigned to for the first time
						   if(dom.numtuples == 0)
						   {
						   	 Domain[] hiddenDomains = getHiddenDoms(relResult);
						   	 if(hiddenDomains.length > 0)
						   	 {
						   	 	if((dom.numattrs + hiddenDomains.length) != (relResult.numattrs + 1))
						   	 		throw new InterpretError("type mismatch between relation " + dom.name +
						   	 		" and relation intended sub relation");
						   	 	else //transform dom to add in the hidden domains
						   	 	{
						   	 		Domain[] newD = new Domain[relResult.numattrs + 1];
						   	 		Domain idd = env.lookupDom(".id", true);
						   	 		newD[0] = idd;
						   	 		for(int k = 0; k < relResult.numattrs; k++)
						   	 			newD[k + 1] = relResult.domains[k];
	
						   	 		Relation newdom = new Relation(dom.name, newD, 0, newD.length, 0, env);
						   	 		newdom.createDataArray(0);
	
						   	 		Environment findEnv = env.removeRel2(dom.name, true);
						   	 		env.put2(newdom, findEnv);
						   	 		dom = env.lookupRel("." + left.name, true);
						   	 	}
						   	 }
						   }
						   
						   //end of addition June 17, 2002
						   
						   dom.insert(surr, relResult);
						} //end of addition June 10, 2002
					}
					else if(id.kind == REL) //for on the fly declaration of relation on LHS (renamed output param)
					{ 
					 //addition by Yi Zheng May 20, 2002
						Relation rel = env.lookupRel(left.name, true);
					  	if (rel == null)
					  		throw new InterpretError(left.name + " as a relation not found");
	 
					 if(stmt.opcode == OP_ASSIGNADD)
					  relResult.assignAdd(null, rel, null);
   					  // bug fix by Zongyan Nov 29, 2004
					  //rel.assignAdd(null, relResult, null);
					 //end of addition May 20, 2002
	
					 	//Relation r = relResult.duplicate(true);  //bug fix by Zongyan Nov 29, 2004
                                                Relation r;
					        if(stmt.opcode == OP_ASSIGNADD)
					 	     r = rel.duplicate(true);
                                                else 
					 	     r = relResult.duplicate(true);
					   	r.name = left.name;
						if (r.data == null) r.load();
					   	((RelInfo) id).rel = r;
					   	env.put(left.name, id); //overwrites the initial entry
					   
					   //added by Yi Zheng March 25, 2002
					   //delete "pat_temp" and save r
					   Interpreter.deleteFile("pat_temp");

					   // added by Zongyan Nov 25,2004
					/*
	  			 	   for(j=0; j<numattrs; j++) {  
		 				if(r.name.compareTo( prefix+domains[j].name) == 0)
                 				break; }
                                           if(j == numattrs) callEnv.put(r);   
					*/
                                           // end of Zongyan's addition Nov 25, 2004
					   r.save();
					}
					else
					{  throw new 
					   InterpretError("Illegal kind for identifier "+left.name);
					}
					break;
				 case COMP :
			//		break; commented out by Yi Zheng March 31, 2002
				 case TEXT :
				 case STMT :
				 case EXPR :
					throw new
					InterpretError("Types 'comp', 'text', 'stmt', and 'expr' " +
								   "are not yet supported in assignment in computations");
				 default :
					throw new InterpretError("Unknown type");
			  }
		   }
		   else if (numChildren == 4)
		   {  //REVISIT: assignment with 4 children
		   //added by Yi Zheng March 24, 2002
		   throw new InterpretError("Assignment with renaming not implemented yet");
		   }
		   else
		   {  throw new  //should never happen
			  InterpretError("Invalid assignment statement");
		   }
		   break;
		case OP_COMMAND :
		//added by Yi Zheng April 8, 2002
		//	interpreter.executeCompCommand(stmt, myEnv);
		interpreter.executeCompCommand(stmt, env); //Yi Zheng April 10, 2002
			break;
		case OP_DECLARATION :
		   switch(stmt.opcode)
		   {  case OP_COMPUTATION :
				 String compName = ((SimpleNode) stmt.jjtGetChild(0)).name;
	                         // weizhong
				 SimpleNode tempNode =(SimpleNode)stmt.jjtGetChild(0);
				 while (tempNode.name == null) {
				   tempNode = (SimpleNode) tempNode.jjtGetChild(0);
				 };
				 compName = tempNode.name;
				 if(env.row == 0)
				 {  //This is our first encounter with this declaration
					//so we must process it.
					Computation c = new Computation(stmt, env);
					env.put(compName, new RelInfo(c));
					
					//Check if this computation is exported
					//(i.e. its name matches that of a formal parameter)
					//If so set the following fields of the SimpleNode stmt:
					//  bits: set bit 10 if comp is exported
					//  Info: store surrogate
					
					if(env.lookupParam(compName, false) != null)
					{ 
						//added by Yi Zheng May 31, 2002
						Environment pEnv = env.getParent();
						if(pEnv != null)
							pEnv.put(compName, new RelInfo(c), true);	//put it in all enclosing parent environments
	
						if (TOP_LEVEL_INOUT) break;
						//end of addition by Yi Zheng May 31, 2002

						
						//Yi Zheng May 14, 2002
						//check if a surrogate already exists by looking up the comptable
						//by this exported computation's fully qualified name
						 String pName = this.parentName.concat("!" + this.name);
						 if (c.parentName.compareTo("") == 0)
						 	c.parentName = pName;
						 String fName = c.parentName.concat("!" + compName);
						 long oldSurr = Global.comptable.lookup(fName);
						 if (oldSurr > 0) //oldSurr exists
						 { ParamInfo pi = env.lookupParam(compName, false);
						   if (pi.type != COMP)
						   	throw new InterpretError("Illegal exportation of " + compName + " as a computation");
						   col = pi.index;
						   ((long[]) pi.env.data[col])[pi.env.row] = oldSurr;
	
						 }
						 else //first 4 lines are original code before May 14, 2002
						 {
						 	stmt.setBit(10); //it is exported
					   		long surr = Surrogate.next();
					   		stmt.setInfo(new Long(surr));
					   		Global.comptable.addEntry(surr, c);
					   		
					   		//Yi Zheng May 14, 2002
					   		Global.comptable.addEntry(fName, new Long(surr));
					   	}
					}
				 } 
				 //If the computation is exported, we store a surrogate
				 //as the value of the corresponding formal parameter
				 if (stmt.isBitSet(10))
				 {  //This computation is exported.
					long surr = ((Long) stmt.getInfo()).longValue();
					ParamInfo pi = env.lookupParam(compName, false);
					//REVISIT: This should be in the semantic checker
					if (pi.type != COMP)
					{  throw new 
					   InterpretError("Illegal exportation of comp"+compName);
					}
					col = pi.index;
					((long[]) pi.env.data[col])[pi.env.row] = surr;
				}else if (env.row > 0)
				{
				//added by Yi Zheng June 21, 2002
					String pName = this.parentName.concat("!" + this.name);
					 
					 String fName = pName.concat("!" + compName);
					 long oldSurr = Global.comptable.lookup(fName);
					 if (oldSurr > 0) //oldSurr exists
					 { ParamInfo pi = env.lookupParam(compName, false);
					   if (pi.type != COMP)
					   	throw new InterpretError("Illegal exportation of " + compName + " as a computation");
					   col = pi.index;
					   ((long[]) pi.env.data[col])[pi.env.row] = oldSurr;
	
					}else throw new InterpretError("surrogate for " + compName + " not found!");
	
				}
			
				 break;
			  case OP_VIEW :{//Yi Zheng June 10, 2002
							SimpleNode init_node = (stmt.isBitSet(0)?(SimpleNode)stmt.jjtGetChild(1):null);
							SimpleNode target_node = null;
							SimpleNode is_node = null;
							
							if(init_node != null)
							{
								is_node = (SimpleNode)stmt.jjtGetChild(2);
								target_node = (stmt.isBitSet(1)?(SimpleNode)stmt.jjtGetChild(3):null);
							}else
							{
								is_node = (SimpleNode)stmt.jjtGetChild(1);
								target_node = (stmt.isBitSet(1)?(SimpleNode)stmt.jjtGetChild(2):null);
							}
							
							transformExpr(is_node, env);
							if(init_node != null) transformExpr(is_node, env);
							if(target_node != null) transformExpr(target_node, env);
							interpreter.executeCompDeclaration(stmt, env);
							break;
						}
			  
			  case OP_LET :{//Yi Zheng June 10, 2002
							SimpleNode be_node = null;
							SimpleNode init_node = (stmt.isBitSet(0)?(SimpleNode)stmt.jjtGetChild(1):null);
							if(init_node != null)
								be_node = (SimpleNode) stmt.jjtGetChild(2);
							else
								be_node = (SimpleNode) stmt.jjtGetChild(1);
							transformExpr(be_node, env);
							if(init_node != null) transformExpr(init_node, env);
							interpreter.executeCompDeclaration(stmt, env);
							break;
						}
	
			  case OP_RELATION :
			  case OP_DOMAIN :
				//addition by Yi Zheng May 20, 2002
			  	 interpreter.executeCompDeclaration(stmt, env);
			  	 break;
			  	//end of addition May 20, 2002
			  default :
				 throw new 
				 InterpretError("Not implemented yet");
		   }
		   break;
		case OP_STATEMENT :
		   switch(opcode) 
		   {  case OP_CONDITIONAL :
				 SimpleNode ifNode, thenNode, elseNode;
				 ifNode = (SimpleNode) stmt.jjtGetChild(0);
				 thenNode = (SimpleNode) stmt.jjtGetChild(1);
		 
				 if (EvalExpr.evalBooleanExpr(ifNode, env))
				 {  runSingleStmt(thenNode, env, null);// last argument added by Zongyan  Nov 25, 2004
				 }
				 else if(numChildren == 3)
				 {  elseNode = (SimpleNode) stmt.jjtGetChild(2);
					runSingleStmt(elseNode, env, null);// last argument added by Zongyan  Nov 25, 2004
				 }
				 break;
			  case OP_STMTBLOCK :
				 for(int stmtNo=0; stmtNo<stmt.jjtGetNumChildren(); stmtNo++)
				 {  runSingleStmt((SimpleNode) stmt.jjtGetChild(stmtNo), env, null);// last argument added by Zongyan  Nov 25, 2004
				 }
				 break;
			  case OP_COMPCALL: //added by Yi Zheng May 20, 2002
			  	interpreter.executeCompStatement(stmt, env);
			  	break;
			  case OP_SEQUENCE :
			  case OP_PARALLEL :
			  case OP_CHOICE :
			  case OP_FORLOOP :
			  case OP_WHILELOOP :
			  case OP_EXIT :
			  case OP_DEADLOCK :
			  case OP_EXEC :
				 throw new 
				 InterpretError("Not implemented yet");
		   }
		   break;
		case OP_UPDATE :
			  //added by Yi Zheng May 20, 2002
			Environment uenv = new Environment(env, env.getTable(), env.reltable, env.domtable);
			
			//added by Yi Zheng June 10, 2002
			switch(stmt.opcode)
			{
				case OP_CHANGE:{
					SimpleNode stmtList = (stmt.isBitSet(0)?(SimpleNode)stmt.jjtGetChild(1):null);
					SimpleNode using = null;
					if(stmtList != null && stmt.isBitSet(1))
						using = (SimpleNode)stmt.jjtGetChild(2);
					else if(stmt.isBitSet(1))
						using = (SimpleNode)stmt.jjtGetChild(1); //impossible
					if(stmtList != null)
					{
						for(int i = 0; i < stmtList.jjtGetNumChildren(); i++)
							transformStmt((SimpleNode)stmtList.jjtGetChild(i), env);
					}
					if(using != null)
					{
						int numChild = using.jjtGetNumChildren();
						if(numChild == 1)
						{
							transformExpr((SimpleNode)using.jjtGetChild(0), env);
						}
						else if(numChild == 3)
						{
							transformExpr((SimpleNode)using.jjtGetChild(2), env);
						}
						else throw new InterpretError("invalid update change");
					}
					break;
				}
				case OP_ADD:
				case OP_DELETE:{
					int numChild = stmt.jjtGetNumChildren();
					if(numChild == 2)
					{
						transformExpr((SimpleNode)stmt.jjtGetChild(1), env);
					}
					else if(numChild == 4)
					{
						transformExpr((SimpleNode)stmt.jjtGetChild(3), env);
					}
					else throw new InterpretError("invalid update add/delete");
					break;
				}
				default:
					throw new InterpretError("invalid update type");
			}
			
			
			//end of addition June 10, 2002
			// produce trigger tree
			int diftrig = interpreter.lookCompUpdate(stmt, uenv, null, uenv.triggerroot);

			// reverse the result -- undo trigger
			interpreter.doCompTrigger(env, uenv.triggerroot, 1);
			if (diftrig == 1){
			  interpreter.executeCompUpdate(stmt, stmt, uenv); 
                        }
			interpreter.RemoveTrigger(uenv.triggerroot);		// clear up
                        // added by Zongyan Nov 25,2004
                        SimpleNode rel_id = (SimpleNode) stmt.jjtGetChild(0);
                        Relation r = null;
                        if(( r = env.lookupRel(rel_id.name, true)) == null){
                              //if(( r = env.lookupNestedRel(rel_id.name, true)) == null){
                                   throw new InterpretError(rel_id.name + " as a relation not found");
                              //}
                        }

                        for(j=0; j<numattrs; j++) {
                              if(r.name.compareTo( prefix+domains[j].name) == 0) break; 
                        }
                        if(j == numattrs){  // not parameter
			      Relation rel = r.duplicate(true);  
                              if(rel.data == null) rel.load();
                              callEnv.put(rel);
                              rel.save();
                        }
                        else{               // parameter
                        }
                        // end of Zongyan's addition Nov 25, 2004

			break;
		  //end of addition May 20, 2002
		default :
		   throw new 
		   InterpretError("Illegal statement");
 		}  /* switch */
	}catch(InterpretError inte)
	{	interpreter.env = oldEnv;
 		throw inte;
	}
 	//addition by Yi Zheng May 20, 2002
 	//restore interpreter.env to oldEnv
 	interpreter.env = oldEnv;
 	//end of addition May 20, 2002
   } /* runSingleStmt() */   
   
   /**
    * get the states in rel
    * @author Yi Zheng
    */
   private Domain[] getHiddenDoms(Relation rel)
   {
   	int num = 0;
   	Vector domVec = new Vector();
   	for(int i = 0; i < rel.numattrs; i++)
   	{
   		if(rel.domains[i].isStateDom == 1)
   		{
   			num++;
   			domVec.addElement(rel.domains[i]);
   		}
   	}
   	
   	Domain[] ret = new Domain[num];
   	if(num > 0)
   		domVec.copyInto(ret);
   	return ret;
   } /* getHiddenDoms */
} /* class Computation */
