

/**
 * Class DBSurrogate implements the algorithm for loading and storing
 * of surrogate value from and to disk.
 *
 * Surrogates generated by the system. They are all different for a
 * database. We use long integers as surragates. Since the largest long
 * integer is about 9.2E+18, it will wrap after 292,471 years, assuming
 * that the surrogates are generated at a speed of 1,000,000 per second.
 *
 * To ensure that the surrogates are uniquely defined over the lifetime
 * of a database , we make it persistent by storing it on disk. Its
 * value is loaded from the disk before using it. Before storing a
 * relation which contains surrogates, we always dump the current value
 * of the surrogate onto disk. 
 * 
 * To ensure the recovering from software or hardware failure when
 * dumping the current value of the surrogate, we use two copies for
 * storing the surrogate.
 *
 * 	A : .surrogate      - current value of the surrogate
 *	B : .surrogate.bak  - last value of the surrogate
 *
 * The following algorithm for loading and storing the value of the
 * surrogate.
 *
 * ------------------------------------------------------------------
 * A	B	load			save
 * ------------------------------------------------------------------
 * no   no	create A, B in order	create A, store current value
 * 		initialize them to 1    create B, store current value
 *		set current value 1
 * no	yes     create A with B.value   as above
 * 		set current value to
 *		B.value
 * yes	no      create B with A.value   create B, store current value
 * 		set current value to    create A, store current value
 *		A.value
 * yes	yes
 *
 *	>	as above		as above
 *
 *	<	create A with B.value   create A, store current value
 *		set current value to    create B, store current value
 *		B.value
 * ------------------------------------------------------------------
 *
 * @author Biao Hao (bhao@cs.mcgill.ca)
 * @version 0.5, 07/02/1997
 */

// THE SIMPLIFIED VERSION OF THE ABOVE ALGORITHM !!!

import java.io.*;


public final class Surrogate implements TypeConstants
{
	/** 
	 * Files in which the value of surrogate is stored.
	 */
//	static int BUF_SIZE = 16 * 1024;

	private static final String fname = ".surrogate";
	/**
	 * Value of surrogate for successive use.
	 */
	private static long value = 1L;

	static {
	try {
	    File f = new File(fname);
			if (f.exists()) {
//		BlockInputStream din = new BlockInputStream(f, BUF_SIZE);
		DataInputStream din = new DataInputStream(new FileInputStream(f));
//		System.out.println("open file" + fname);
//		value = din.readLong();

		// modified by Weizhong Sun
		// read Ascii
		Relation r = new Relation();
		String v = r.strRead(din, STRING);
//		System.out.println("the string:" + v);
		Long vv = new Long(v);
		value = vv.longValue();
		
		din.close();
	    }
	} catch (IOException e) {
	    System.out.println("surrogate initialization error");
	    value = Long.MAX_VALUE%2; save();
	} 
	}



	// Test the Surrogate class. Also used to set surrogate value.
	public static void main(String[] args)
	{
	if (args.length == 2 && args[0].equals("-setSurrogate")) {
	    try {
		long val = Long.parseLong(args[1]);
		value = val; save();
		System.out.println("set surrogate to : " + value);
	    } catch (NumberFormatException e) {
		System.out.println("NumberFormatException, try again");
	    }
	} else {
			System.out.println(Surrogate.next());
	    Surrogate.save();
	}
	}
	/**
	 * The next value of surrogate.
	 */
	public static long next()
	{
		return value++;
	}
	/**
	 * Save the surrogate value onto disk file .surrogate.
	 */
	public static void save()
	{
//	BlockOutputStream dout = null;
	DataOutputStream dout = null;
		try {
//			dout = new BlockOutputStream(fname, BUF_SIZE);
			dout = new DataOutputStream(new FileOutputStream(fname));
//			dout.writeLong(value);

			// modified by Weizhong Sun
			// save as Ascii
			String v = new String("");
			v = v.valueOf(value);
			dout.writeBytes(v);
//			dout.writeBytes("^F");
			dout.writeByte(6);

			dout.close();
		} catch (IOException e) {
	    System.out.println("surrogate : " + value + " save error");
		} 
	}
}
